import itertools
import logging
import warnings
from collections import defaultdict
from typing import Any

import pandas as pd
from pydantic import ValidationError

from cognite.neat.rules.models import SchemaCompleteness
from cognite.neat.rules.models._rdfpath import RDFPath
from cognite.neat.rules.models.entities import ClassEntity, EntityTypes, ParentClassEntity, ReferenceEntity
from cognite.neat.rules.models.information import InformationClass, InformationProperty, InformationRules
from cognite.neat.utils.utils import get_inheritance_path

from ._base import BaseAnalysis


class InformationArchitectRulesAnalysis(BaseAnalysis):
    """Assumes analysis over only the complete schema"""

    def __init__(self, rules: InformationRules):
        self.rules = rules

    @property
    def directly_referred_classes(self) -> set[ClassEntity]:
        return {
            class_.reference.as_class_entity()
            for class_ in self.rules.classes
            if self.rules.reference
            and class_.reference
            and isinstance(class_.reference, ReferenceEntity)
            and class_.reference.prefix == self.rules.reference.metadata.prefix
        }

    @property
    def inherited_referred_classes(self) -> set[ClassEntity]:
        dir_referred_classes = self.directly_referred_classes
        inherited_referred_classes = []
        for class_ in dir_referred_classes:
            inherited_referred_classes.extend(self.class_inheritance_path(class_))
        return set(inherited_referred_classes)

    def class_parent_pairs(self) -> dict[ClassEntity, list[ParentClassEntity]]:
        """This only returns class - parent pairs only if parent is in the same data model"""
        class_subclass_pairs: dict[ClassEntity, list[ParentClassEntity]] = {}

        if not self.rules:
            return class_subclass_pairs

        for definition in self.rules.classes:
            class_subclass_pairs[definition.class_] = []

            if definition.parent is None:
                continue

            for parent in definition.parent:
                if parent.prefix == definition.class_.prefix:
                    class_subclass_pairs[definition.class_].append(parent)
                else:
                    warnings.warn(
                        f"Parent class {parent} of class {definition} is not in the same namespace, skipping !",
                        stacklevel=2,
                    )

        return class_subclass_pairs

    def classes_with_properties(
        self, consider_inheritance: bool = False
    ) -> dict[ClassEntity, list[InformationProperty]]:
        """Returns classes that have been defined in the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Dictionary of classes with a list of properties defined for them

        !!! note "consider_inheritance"
            If consider_inheritance is True, properties from parent classes will also be considered.
            This means if a class has a parent class, and the parent class has properties defined for it,
            while we do not have any properties defined for the child class, we will still consider the
            properties from the parent class. If consider_inheritance is False, we will only consider
            properties defined for the child class, thus if no properties are defined for the child class,
            it will not be included in the returned dictionary.
        """

        class_property_pairs: dict[ClassEntity, list[InformationProperty]] = defaultdict(list)

        for property_ in self.rules.properties:
            class_property_pairs[property_.class_].append(property_)

        if consider_inheritance:
            class_parent_pairs = self.class_parent_pairs()
            for class_ in class_parent_pairs:
                self._add_inherited_properties(class_, class_property_pairs, class_parent_pairs)

        return class_property_pairs

    def class_inheritance_path(self, class_: ClassEntity | str) -> list[ClassEntity]:
        class_ = class_ if isinstance(class_, ClassEntity) else ClassEntity.load(class_)
        class_parent_pairs = self.class_parent_pairs()
        return get_inheritance_path(class_, class_parent_pairs)

    @classmethod
    def _add_inherited_properties(
        cls,
        class_: ClassEntity,
        class_property_pairs: dict[ClassEntity, list[InformationProperty]],
        class_parent_pairs: dict[ClassEntity, list[ParentClassEntity]],
    ):
        inheritance_path = get_inheritance_path(class_, class_parent_pairs)
        for parent in inheritance_path:
            # ParentClassEntity -> ClassEntity to match the type of class_property_pairs
            if parent.as_class_entity() in class_property_pairs:
                for property_ in class_property_pairs[parent.as_class_entity()]:
                    property_ = property_.model_copy()

                    # This corresponds to importing properties from parent class
                    # making sure that the property is attached to desired child class
                    property_.class_ = class_
                    property_.inherited = True

                    # need same if we have RDF path to make sure that the starting class is the

                    if class_ in class_property_pairs:
                        class_property_pairs[class_].append(property_)
                    else:
                        class_property_pairs[class_] = [property_]

    def class_property_pairs(
        self, only_rdfpath: bool = False, consider_inheritance: bool = False
    ) -> dict[ClassEntity, dict[str, InformationProperty]]:
        """Returns a dictionary of classes with a dictionary of properties associated with them.

        Args:
            only_rdfpath : To consider only properties which have rule `rdfpath` set. Defaults False
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Dictionary of classes with a dictionary of properties associated with them.

        !!! note "difference to get_classes_with_properties"
            This method returns a dictionary of classes with a dictionary of properties associated with them.
            While get_classes_with_properties returns a dictionary of classes with a list of
            properties defined for them,
            here we filter the properties based on the `only_rdfpath` parameter and only consider
            the first definition of a property if it is defined more than once.

        !!! note "only_rdfpath"
            If only_rdfpath is True, only properties with RuleType.rdfpath will be returned as
            a part of the dictionary of properties related to a class. Otherwise, all properties
            will be returned.

        !!! note "consider_inheritance"
            If consider_inheritance is True, properties from parent classes will also be considered.
            This means if a class has a parent class, and the parent class has properties defined for it,
            while we do not have any properties defined for the child class, we will still consider the
            properties from the parent class. If consider_inheritance is False, we will only consider
            properties defined for the child class, thus if no properties are defined for the child class,
            it will not be included in the returned dictionary.
        """
        # TODO: https://cognitedata.atlassian.net/jira/software/projects/NEAT/boards/893?selectedIssue=NEAT-78

        class_property_pairs = {}

        for class_, properties in self.classes_with_properties(consider_inheritance).items():
            processed_properties = {}
            for property_ in properties:
                if property_.property_ in processed_properties:
                    # TODO: use appropriate Warning class from _exceptions.py
                    # if missing make one !
                    warnings.warn(
                        f"Property {property_.property_} for {class_} has been defined more than once!"
                        " Only the first definition will be considered, skipping the rest..",
                        stacklevel=2,
                    )
                    continue

                if (only_rdfpath and isinstance(property_.transformation, RDFPath)) or not only_rdfpath:
                    processed_properties[property_.property_] = property_
            class_property_pairs[class_] = processed_properties

        return class_property_pairs

    def class_linkage(self, consider_inheritance: bool = False) -> pd.DataFrame:
        """Returns a dataframe with the class linkage of the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Dataframe with the class linkage of the data model
        """

        class_linkage = pd.DataFrame(columns=["source_class", "target_class", "connecting_property", "max_occurrence"])

        class_property_pairs = self.classes_with_properties(consider_inheritance)
        properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

        for property_ in properties:
            if property_.type_ == EntityTypes.object_property:
                new_row = pd.Series(
                    {
                        "source_class": property_.class_,
                        "connecting_property": property_.property_,
                        "target_class": property_.value_type,
                        "max_occurrence": property_.max_count,
                    }
                )
                class_linkage = pd.concat([class_linkage, new_row.to_frame().T], ignore_index=True)

        class_linkage.drop_duplicates(inplace=True)
        class_linkage = class_linkage[["source_class", "connecting_property", "target_class", "max_occurrence"]]

        return class_linkage

    def connected_classes(self, consider_inheritance: bool = False) -> set[ClassEntity]:
        """Return a set of classes that are connected to other classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that are connected to other classes
        """
        class_linkage = self.class_linkage(consider_inheritance)
        return set(class_linkage.source_class.values).union(set(class_linkage.target_class.values))

    def defined_classes(self, consider_inheritance: bool = False) -> set[ClassEntity]:
        """Returns classes that have properties defined for them in the data model.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that have been defined in the data model
        """
        class_property_pairs = self.classes_with_properties(consider_inheritance)
        properties = list(itertools.chain.from_iterable(class_property_pairs.values()))

        return {property.class_ for property in properties}

    def disconnected_classes(self, consider_inheritance: bool = False) -> set[ClassEntity]:
        """Return a set of classes that are disconnected (i.e. isolated) from other classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that are disconnected from other classes
        """
        return self.defined_classes(consider_inheritance) - self.connected_classes(consider_inheritance)

    def symmetrically_connected_classes(
        self, consider_inheritance: bool = False
    ) -> set[tuple[ClassEntity, ClassEntity]]:
        """Returns a set of pairs of symmetrically linked classes.

        Args:
            consider_inheritance: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of pairs of symmetrically linked classes

        !!! note "Symmetrically Connected Classes"
            Symmetrically connected classes are classes that are connected to each other
            in both directions. For example, if class A is connected to class B, and class B
            is connected to class A, then classes A and B are symmetrically connected.
        """

        # TODO: Find better name for this method
        sym_pairs: set[tuple[ClassEntity, ClassEntity]] = set()

        class_linkage = self.class_linkage(consider_inheritance)
        if class_linkage.empty:
            return sym_pairs

        for _, row in class_linkage.iterrows():
            source = row.source_class
            target = row.target_class
            target_targets = class_linkage[class_linkage.source_class == target].target_class.values
            if source in target_targets and (source, target) not in sym_pairs:
                sym_pairs.add((source, target))
        return sym_pairs

    def as_property_dict(
        self,
    ) -> dict[str, list[InformationProperty]]:
        """This is used to capture all definitions of a property in the data model."""
        property_dict: dict[str, list[InformationProperty]] = defaultdict(list)
        for definition in self.rules.properties:
            property_dict[definition.property_].append(definition)
        return property_dict

    def as_class_dict(self) -> dict[str, InformationClass]:
        """This is to simplify access to classes through dict."""
        class_dict: dict[str, InformationClass] = {}
        for definition in self.rules.classes:
            class_dict[str(definition.class_.suffix)] = definition
        return class_dict

    def subset_rules(self, desired_classes: set[ClassEntity]) -> InformationRules:
        """
        Subset rules to only include desired classes and their properties.

        Args:
            desired_classes: Desired classes to include in the reduced data model

        Returns:
            Instance of InformationRules

        !!! note "Inheritance"
            If desired classes contain a class that is a subclass of another class(es), the parent class(es)
            will be included in the reduced data model as well even though the parent class(es) are
            not in the desired classes set. This is to ensure that the reduced data model is
            consistent and complete.

        !!! note "Partial Reduction"
            This method does not perform checks if classes that are value types of desired classes
            properties are part of desired classes. If a class is not part of desired classes, but it
            is a value type of a property of a class that is part of desired classes, derived reduced
            rules will be marked as partial.

        !!! note "Validation"
            This method will attempt to validate the reduced rules with custom validations.
            If it fails, it will return a partial rules with a warning message, validated
            only with base Pydantic validators.
        """

        if self.rules.metadata.schema_ is not SchemaCompleteness.complete:
            raise ValueError("Rules are not complete cannot perform reduction!")
        class_as_dict = self.as_class_dict()
        class_parents_pairs = self.class_parent_pairs()
        defined_classes = self.defined_classes(consider_inheritance=True)

        possible_classes = defined_classes.intersection(desired_classes)
        impossible_classes = desired_classes - possible_classes

        # need to add all the parent classes of the desired classes to the possible classes
        parents: set[ClassEntity] = set()
        for class_ in possible_classes:
            parents = parents.union(
                {parent.as_class_entity() for parent in get_inheritance_path(class_, class_parents_pairs)}
            )
        possible_classes = possible_classes.union(parents)

        if not possible_classes:
            logging.error("None of the desired classes are defined in the data model!")
            raise ValueError("None of the desired classes are defined in the data model!")

        if impossible_classes:
            logging.warning(f"Could not find the following classes defined in the data model: {impossible_classes}")
            warnings.warn(
                f"Could not find the following classes defined in the data model: {impossible_classes}", stacklevel=2
            )

        reduced_data_model: dict[str, Any] = {
            "metadata": self.rules.metadata.model_copy(),
            "prefixes": (self.rules.prefixes or {}).copy(),
            "classes": [],
            "properties": [],
        }

        logging.info(f"Reducing data model to only include the following classes: {possible_classes}")
        for class_ in possible_classes:
            reduced_data_model["classes"].append(class_as_dict[str(class_.suffix)])

        class_property_pairs = self.classes_with_properties(consider_inheritance=False)

        for class_, properties in class_property_pairs.items():
            if class_ in possible_classes:
                reduced_data_model["properties"].extend(properties)

        try:
            return InformationRules(**reduced_data_model)
        except ValidationError as e:
            warnings.warn(f"Reduced data model is not complete: {e}", stacklevel=2)
            reduced_data_model["metadata"].schema_ = SchemaCompleteness.partial
            return InformationRules.model_construct(**reduced_data_model)



import xml.etree.ElementTree as ET
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from xml.etree.ElementTree import Element

from rdflib import RDF, RDFS, XSD, Literal, Namespace, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import as_neat_compliant_uri
from cognite.neat.utils.xml_ import get_children, iterate_tree

DEXPI = Namespace("http://sandbox.dexpi.org/rdl/")
QUDT = Namespace("https://qudt.org/vocab/unit/")


class DexpiExtractor(BaseExtractor):
    """
    DEXPI-XML extractor of RDF triples

    Args:
        root: XML root element of DEXPI file.
        namespace: Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to DEFAULT_NAMESPACE.
    """

    def __init__(
        self,
        root: Element,
        namespace: Namespace | None = None,
    ):
        self.root = root
        self.namespace = namespace or DEFAULT_NAMESPACE

    @classmethod
    def from_file(cls, filepath: str | Path, namespace: Namespace | None = None):
        return cls(ET.parse(filepath).getroot(), namespace)

    @classmethod
    def from_url(cls, url: str, namespace: Namespace | None = None):
        from io import BytesIO

        import requests

        response = requests.get(url)
        response.raise_for_status()
        return cls(ET.parse(BytesIO(response.content)).getroot(), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extracts RDF triples from DEXPI XML file."""

        for element in iterate_tree(self.root):
            yield from self._element2triples(element, self.namespace)

    @classmethod
    def _element2triples(cls, element: Element, namespace: Namespace) -> list[Triple]:
        """Converts an element to triples."""
        triples: list[Triple] = []

        if (
            "ComponentClass" in element.attrib
            and element.attrib["ComponentClass"] != "Label"
            and "ID" in element.attrib
        ):
            id_ = namespace[element.attrib["ID"]]

            if node_triples := cls._element2node_triples(id_, element):
                triples.extend(node_triples)

            if edge_triples := cls._element2edge_triples(id_, element, namespace):
                triples.extend(edge_triples)

        return triples

    @classmethod
    def _element2edge_triples(cls, id_: URIRef, element: Element, namespace: Namespace) -> list[Triple]:
        triples: list[Triple] = []

        # connection triples
        if connections := get_children(element, "Connection"):
            for connection in connections:
                if "FromID" in connection.attrib and "ToID" in connection.attrib:
                    triples.append(
                        (
                            namespace[connection.attrib["FromID"]],
                            DEXPI.connection,
                            namespace[connection.attrib["ToID"]],
                        )
                    )

        # association triples
        if associations := get_children(element, "Association"):
            for association in associations:
                if "Type" in association.attrib and "ItemID" in association.attrib:
                    association_type = cls._to_uri_friendly_association_type(association)

                    triples.append(
                        (
                            id_,
                            DEXPI[f"association/{association_type}"],
                            namespace[association.attrib["ItemID"]],
                        )
                    )

        # children-parent triples
        for child in element:
            if "ID" in child.attrib and child.tag != "Label":
                camel_case_property = child.tag[0].lower() + child.tag[1:]
                triples.append(
                    (
                        id_,
                        DEXPI[f"children/{camel_case_property}"],
                        namespace[child.attrib["ID"]],
                    )
                )

        return triples

    @classmethod
    def _to_uri_friendly_association_type(cls, association: Element):
        association_type = "".join(
            [word.capitalize() if i != 0 else word for i, word in enumerate(association.attrib["Type"].split(" "))]
        )

        return association_type

    @classmethod
    def _element2node_triples(cls, id_: URIRef, element: Element) -> list[Triple]:
        """Converts an XML element to triples."""
        triples: list[Triple] = []

        # setting these to None this is the order of getting the type
        component_class: str | None = None
        component_name: str | None = None
        tag: str | None = None

        # adding tag triple if exists
        if tag := element.tag:
            triples.append((id_, DEXPI.tag, Literal(str(tag))))

        # adding attributes triples
        if attributes := element.attrib:
            if component_class := attributes.get("ComponentClass", None):
                triples.append((id_, DEXPI.ComponentClass, Literal(component_class)))
            if component_name := attributes.get("ComponentName", None):
                triples.append((id_, DEXPI.ComponentName, Literal(component_name)))
            if component_class_uri := attributes.get("ComponentClassURI", None):
                triples.append((id_, DEXPI.ComponentClassURI, URIRef(component_class_uri)))

        triples.append(
            (
                id_,
                RDF.type,
                as_neat_compliant_uri(DEFAULT_NAMESPACE[component_class or component_name or tag or "Unknown"]),
            )
        )

        # add label triple
        if label := cls._get_element_label(element):
            triples.append((id_, RDFS.label, Literal(label)))

        # add generic attributes triples
        if generic_attributes := cls._get_element_generic_attributes(element):
            for attribute, value_definitions in generic_attributes.items():
                predicate = as_neat_compliant_uri(attribute)
                for value_definition in value_definitions:
                    if literal := cls._value_definition2literal(value_definition):
                        triples.append((id_, predicate, literal))

        return triples

    @classmethod
    def _value_definition2literal(cls, definition: dict, make_unit_datatype: bool = False) -> Literal | None:
        if "Value" not in definition or "Format" not in definition:
            return None

        if "Units" in definition and "Value" in definition:
            if make_unit_datatype and "UnitsURI" in definition:
                return Literal(definition["Value"], datatype=URIRef(definition["UnitsURI"]))

            else:
                return Literal(definition["Value"], datatype=XSD.float)

        # case: when language is present we create add language tag to the literal
        elif "Language" in definition and "Value" in definition:
            return Literal(definition["Value"], lang=definition["Language"])

        # case: when ValueURI is present we use it instead of Value
        # this would be candidate for ENUMs in CDF
        elif "ValueURI" in definition:
            return Literal(definition["ValueURI"], datatype=XSD[definition["Format"]])

        # case: when Format is not string we make sure to add the datatype
        elif definition["Format"].lower() != "string":
            return Literal(definition["Value"], datatype=XSD[definition["Format"]])

        # case: when Format is string we add the literal without datatype (easier to read triples, less noise)
        else:
            return Literal(definition["Value"])

    @classmethod
    def _get_element_label(cls, element: Element) -> str | None:
        if children := get_children(element, "Label", 1):
            if grandchildren := get_children(children[0], "Text", 1):
                if "String" in grandchildren[0].attrib:
                    return grandchildren[0].attrib["String"]

        # extension for schema version 3.3, where text is used to "label" without a <label> parent
        elif children := get_children(element, "Text", 1):
            if "String" in children[0].attrib:
                return children[0].attrib["String"]

        return None

    @classmethod
    def _get_element_generic_attributes(cls, element: Element) -> dict:
        # TODO: This requires more work as there are multiple groupings of GenericAttributes

        attributes = defaultdict(list)
        if children := get_children(element, "GenericAttributes", 1):
            if grandchildren := get_children(children[0], "GenericAttribute"):
                for generic_attribute in grandchildren:
                    # extension for schema version 3.3, where "AttributeURI" is not included
                    if name := generic_attribute.attrib.get("Name", None):
                        attribute_uri = as_neat_compliant_uri(DEFAULT_NAMESPACE[name])
                        if attribute_uri not in attributes:
                            attributes[attribute_uri] = [generic_attribute.attrib]
                        else:
                            attributes[attribute_uri].append(generic_attribute.attrib)

        return attributes



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, cast, overload

from cognite.neat.rules.models._base import _add_alias
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.entities import (
    ClassEntity,
    MultiValueTypeInfo,
    Unknown,
    UnknownEntity,
)
from cognite.neat.rules.models.information._rules_input import InformationClassInput, InformationMetadataInput

from ._rules import AssetProperty, AssetRules


@dataclass
class AssetMetadataInput(InformationMetadataInput): ...


@dataclass
class AssetPropertyInput:
    class_: str
    property_: str
    value_type: str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    reference: str | None = None
    match_type: str | None = None
    transformation: str | None = None
    implementation: str | None = None

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "AssetPropertyInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["AssetPropertyInput"]: ...

    @classmethod
    def load(
        cls, data: dict[str, Any] | list[dict[str, Any]] | None
    ) -> "AssetPropertyInput | list[AssetPropertyInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]

        _add_alias(data, AssetProperty)
        return cls(
            class_=data.get("class_"),  # type: ignore[arg-type]
            property_=data.get("property_"),  # type: ignore[arg-type]
            name=data.get("name", None),
            description=data.get("description", None),
            comment=data.get("comment", None),
            value_type=data.get("value_type"),  # type: ignore[arg-type]
            min_count=data.get("min_count", None),
            max_count=data.get("max_count", None),
            default=data.get("default", None),
            reference=data.get("reference", None),
            match_type=data.get("match_type", None),
            transformation=data.get("transformation", None),
            implementation=data.get("implementation", None),
        )

    def dump(self, default_prefix: str) -> dict[str, Any]:
        value_type: MultiValueTypeInfo | DataType | ClassEntity | UnknownEntity

        # property holding xsd data type
        # check if it is multi value type
        if "|" in self.value_type:
            value_type = MultiValueTypeInfo.load(self.value_type)
            value_type.set_default_prefix(default_prefix)

        elif DataType.is_data_type(self.value_type):
            value_type = DataType.load(self.value_type)

        # unknown value type
        elif self.value_type == str(Unknown):
            value_type = UnknownEntity()

        # property holding link to class
        else:
            value_type = ClassEntity.load(self.value_type, prefix=default_prefix)

        return {
            "Class": ClassEntity.load(self.class_, prefix=default_prefix),
            "Property": self.property_,
            "Name": self.name,
            "Description": self.description,
            "Comment": self.comment,
            "Value Type": value_type,
            "Min Count": self.min_count,
            "Max Count": self.max_count,
            "Default": self.default,
            "Reference": self.reference,
            "Match Type": self.match_type,
            "Transformation": self.transformation,
            "Implementation": self.implementation,
        }


class AssetClassInput(InformationClassInput): ...


@dataclass
class AssetRulesInput:
    metadata: AssetMetadataInput
    properties: Sequence[AssetPropertyInput]
    classes: Sequence[AssetClassInput]
    last: "AssetRulesInput | AssetRules | None" = None
    reference: "AssetRulesInput | AssetRules | None" = None

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "AssetRulesInput": ...

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    def load(cls, data: dict | None) -> "AssetRulesInput | None":
        if data is None:
            return None
        _add_alias(data, AssetRules)

        return cls(
            metadata=AssetMetadataInput.load(data.get("metadata")),  # type: ignore[arg-type]
            properties=AssetPropertyInput.load(data.get("properties")),  # type: ignore[arg-type]
            classes=InformationClassInput.load(data.get("classes")),  # type: ignore[arg-type]
            last=AssetRulesInput.load(data.get("last")),
            reference=AssetRulesInput.load(data.get("reference")),
        )

    def as_rules(self) -> AssetRules:
        return AssetRules.model_validate(self.dump())

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix
        reference: dict[str, Any] | None = None
        if isinstance(self.reference, AssetRulesInput):
            reference = self.reference.dump()
        elif isinstance(self.reference, AssetRules):
            # We need to load through the AssetRulesInput to set the correct default space and version
            reference = AssetRulesInput.load(self.reference.model_dump()).dump()
        last: dict[str, Any] | None = None
        if isinstance(self.last, AssetRulesInput):
            last = self.last.dump()
        elif isinstance(self.last, AssetRules):
            # We need to load through the AssetRulesInput to set the correct default space and version
            last = AssetRulesInput.load(self.last.model_dump()).dump()

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Classes=[class_.dump(default_prefix) for class_ in self.classes],
            Last=last,
            Reference=reference,
        )



from ._classic_cdf import (
    AddAssetDepth,
    AssetEventConnector,
    AssetFileConnector,
    AssetRelationshipConnector,
    AssetSequenceConnector,
    AssetTimeSeriesConnector,
)

__all__ = [
    "AddAssetDepth",
    "AssetTimeSeriesConnector",
    "AssetSequenceConnector",
    "AssetFileConnector",
    "AssetEventConnector",
    "AssetRelationshipConnector",
]

Transformers = (
    AddAssetDepth
    | AssetTimeSeriesConnector
    | AssetSequenceConnector
    | AssetFileConnector
    | AssetEventConnector
    | AssetRelationshipConnector
)



import json
import logging
import os
import shutil
import sys
from pathlib import Path
from typing import Literal, cast

import yaml
from pydantic import BaseModel, Field
from yaml import safe_load

from cognite.neat.constants import EXAMPLE_GRAPHS, EXAMPLE_RULES, EXAMPLE_WORKFLOWS
from cognite.neat.utils.cdf import InteractiveCogniteClient, ServiceCogniteClient

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self

LOG_FORMAT = "%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class RulesStoreType(StrEnum):
    """Rules Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"
    GOOGLE_SHEET = "google_sheet"


class WorkflowsStoreType(StrEnum):
    """Workflows Store type"""

    CDF = "cdf"
    FILE = "file"
    URL = "url"


class Config(BaseModel):
    workflows_store_type: WorkflowsStoreType = WorkflowsStoreType.FILE
    data_store_path: Path = Field(default_factory=lambda: Path.cwd() / "data")

    workflow_downloader_filter: list[str] | None = Field(
        description="List of workflow names+tags to filter on when downloading workflows from CDF. "
        "Example name:workflow_name=version,tag:tag_name",
        default=None,
    )

    cdf_client: InteractiveCogniteClient | ServiceCogniteClient = ServiceCogniteClient()
    cdf_default_dataset_id: int = 0
    load_examples: bool = True

    log_level: Literal["ERROR", "WARNING", "INFO", "DEBUG"] = "INFO"
    log_format: str = LOG_FORMAT
    download_workflows_from_cdf: bool = Field(
        default=False, description="Downloads all workflows from CDF automatically and stores them locally"
    )

    stop_on_error: bool = False

    @property
    def _dir_suffix(self) -> str:
        is_test_running = "pytest" in sys.modules
        if is_test_running:
            # Todo change the below to f"-{os.getpid()}" when all tests supports parallel execution.
            return ""
        return ""

    @property
    def rules_store_path(self) -> Path:
        return self.data_store_path / f"rules{self._dir_suffix}"

    @property
    def workflows_store_path(self) -> Path:
        return self.data_store_path / f"workflows{self._dir_suffix}"

    @property
    def source_graph_path(self) -> Path:
        return self.data_store_path / f"source-graphs{self._dir_suffix}"

    @property
    def staging_path(self) -> Path:
        return self.data_store_path / f"staging{self._dir_suffix}"

    @classmethod
    def from_yaml(cls, filepath: Path) -> Self:
        return cls(**safe_load(filepath.read_text()))

    def to_yaml(self, filepath: Path):
        # Parse as json to avoid Path and Enum objects
        dump = json.loads(self.json())

        with filepath.open("w") as f:
            yaml.safe_dump(dump, f)

    @classmethod
    def from_env(cls) -> Self:
        missing = "Missing"
        cdf_config = ServiceCogniteClient(
            project=os.environ.get("NEAT_CDF_PROJECT", missing),
            client_id=os.environ.get("NEAT_CDF_CLIENT_ID", missing),
            client_secret=os.environ.get("NEAT_CDF_CLIENT_SECRET", missing),
            base_url=os.environ.get("NEAT_CDF_BASE_URL", missing),
            token_url=os.environ.get("NEAT_CDF_TOKEN_URL", missing),
            scopes=[os.environ.get("NEAT_CDF_SCOPES", missing)],
            timeout=int(os.environ.get("NEAT_CDF_CLIENT_TIMEOUT", "60")),
            max_workers=int(os.environ.get("NEAT_CDF_CLIENT_MAX_WORKERS", "3")),
        )

        if workflow_downloader_filter_value := os.environ.get("NEAT_WORKFLOW_DOWNLOADER_FILTER", None):
            workflow_downloader_filter = workflow_downloader_filter_value.split(",")
        else:
            workflow_downloader_filter = None

        return cls(
            cdf_client=cdf_config,
            workflows_store_type=os.environ.get(  # type: ignore[arg-type]
                "NEAT_WORKFLOWS_STORE_TYPE", WorkflowsStoreType.FILE
            ),
            data_store_path=Path(os.environ.get("NEAT_DATA_PATH", "/app/data")),
            cdf_default_dataset_id=int(os.environ.get("NEAT_CDF_DEFAULT_DATASET_ID", 6476640149881990)),
            log_level=cast(Literal["ERROR", "WARNING", "INFO", "DEBUG"], os.environ.get("NEAT_LOG_LEVEL", "INFO")),
            workflow_downloader_filter=workflow_downloader_filter,
            load_examples=bool(os.environ.get("NEAT_LOAD_EXAMPLES", True) in ["True", "true", "1"]),
        )


def copy_examples_to_directory(config: Config):
    """
    Copier over all the examples to the target_data_directory,
    without overwriting

    Args:
        target_data_dir : The target directory
        suffix : The suffix to add to the directory names

    """

    print(f"Copying examples into {config.data_store_path}")
    _copy_examples(EXAMPLE_RULES, config.rules_store_path)
    _copy_examples(EXAMPLE_GRAPHS, config.source_graph_path)
    _copy_examples(EXAMPLE_WORKFLOWS, config.workflows_store_path)
    config.staging_path.mkdir(exist_ok=True, parents=True)


def create_data_dir_structure(config: Config) -> None:
    """
    Create the data directory structure in empty directory

    Args:
        target_data_dir : The target directory
        suffix : The suffix to add to the directory names

    """
    for path in (
        config.rules_store_path,
        config.source_graph_path,
        config.staging_path,
        config.workflows_store_path,
    ):
        path.mkdir(exist_ok=True, parents=True)


def _copy_examples(source_dir: Path, target_dir: Path):
    for current in source_dir.rglob("*"):
        if current.is_dir():
            continue
        relative = current.relative_to(source_dir)
        if not (target := target_dir / relative).exists():
            target.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy2(current, target)


def configure_logging(level: str = "DEBUG", log_format: str = LOG_FORMAT):
    """Configure logging based on config."""
    logging.basicConfig(format=log_format, level=logging.getLevelName(level), datefmt=LOG_DATE_FORMAT)



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Event, EventList
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import string_to_ideal_type


class EventsExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusions Events into Neat.

    Args:
        events (Iterable[Event]): An iterable of events.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
    """

    def __init__(
        self,
        events: Iterable[Event],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.events = events

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(cast(Iterable[Event], client.events(data_set_external_ids=data_set_external_id)), namespace)

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(EventList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extract events as triples."""
        for event in self.events:
            yield from self._event2triples(event, self.namespace)

    @classmethod
    def _event2triples(cls, event: Event, namespace: Namespace) -> list[Triple]:
        id_ = namespace[f"Event_{event.id}"]

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, namespace.Event)]

        # Create attributes

        if event.external_id:
            triples.append((id_, namespace.external_id, Literal(event.external_id)))

        if event.source:
            triples.append((id_, namespace.type, Literal(event.source)))

        if event.type:
            triples.append((id_, namespace.type, Literal(event.type)))

        if event.subtype:
            triples.append((id_, namespace.subtype, Literal(event.subtype)))

        if event.metadata:
            for key, value in event.metadata.items():
                if value:
                    type_aware_value = string_to_ideal_type(value)
                    try:
                        triples.append((id_, namespace[key], URIRef(str(AnyHttpUrl(type_aware_value)))))  # type: ignore
                    except ValidationError:
                        triples.append((id_, namespace[key], Literal(type_aware_value)))

        if event.description:
            triples.append((id_, namespace.description, Literal(event.description)))

        if event.created_time:
            triples.append(
                (id_, namespace.created_time, Literal(datetime.fromtimestamp(event.created_time / 1000, timezone.utc)))
            )

        if event.last_updated_time:
            triples.append(
                (
                    id_,
                    namespace.last_updated_time,
                    Literal(datetime.fromtimestamp(event.last_updated_time / 1000, timezone.utc)),
                )
            )

        if event.start_time:
            triples.append(
                (
                    id_,
                    namespace.start_time,
                    Literal(datetime.fromtimestamp(event.start_time / 1000, timezone.utc)),
                )
            )

        if event.end_time:
            triples.append(
                (
                    id_,
                    namespace.end_time,
                    Literal(datetime.fromtimestamp(event.end_time / 1000, timezone.utc)),
                )
            )

        if event.data_set_id:
            triples.append((id_, namespace.data_set_id, namespace[f"Dataset_{event.data_set_id}"]))

        if event.asset_ids:
            for asset_id in event.asset_ids:
                triples.append((id_, namespace.asset, namespace[f"Asset_{asset_id}"]))

        return triples



from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetList
from rdflib import RDF, Literal, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE
from cognite.neat.graph.extractors._base import BaseExtractor
from cognite.neat.graph.models import Triple
from cognite.neat.utils.utils import create_sha256_hash, string_to_ideal_type


class AssetsExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusions Assets into Neat.

    Args:
        assets (Iterable[Asset]): An iterable of assets.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
    """

    def __init__(
        self,
        assets: Iterable[Asset],
        namespace: Namespace | None = None,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.assets = assets

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
    ):
        return cls(cast(Iterable[Asset], client.assets(data_set_external_ids=data_set_external_id)), namespace)

    @classmethod
    def from_hierarchy(cls, client: CogniteClient, root_asset_external_id: str, namespace: Namespace | None = None):
        return cls(cast(Iterable[Asset], client.assets(asset_subtree_external_ids=root_asset_external_id)), namespace)

    @classmethod
    def from_file(cls, file_path: str, namespace: Namespace | None = None):
        return cls(AssetList.load(Path(file_path).read_text()), namespace)

    def extract(self) -> Iterable[Triple]:
        """Extracts an asset with the given asset_id."""
        for asset in self.assets:
            yield from self._asset2triples(asset, self.namespace)

    @classmethod
    def _asset2triples(cls, asset: Asset, namespace: Namespace) -> list[Triple]:
        """Converts an asset to triples."""
        id_ = namespace[f"Asset_{asset.id}"]

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, namespace["Asset"])]

        # Create attributes
        if asset.name:
            triples.append((id_, namespace.name, Literal(asset.name)))

        if asset.description:
            triples.append((id_, namespace.description, Literal(asset.description)))

        if asset.external_id:
            triples.append((id_, namespace.external_id, Literal(asset.external_id)))

        if asset.source:
            triples.append((id_, namespace.source, Literal(asset.source)))

        # properties ref creation and update
        triples.append(
            (
                id_,
                namespace.created_time,
                Literal(datetime.fromtimestamp(asset.created_time / 1000, timezone.utc)),
            )
        )
        triples.append(
            (
                id_,
                namespace.last_updated_time,
                Literal(datetime.fromtimestamp(asset.last_updated_time / 1000, timezone.utc)),
            )
        )

        if asset.labels:
            for label in asset.labels:
                # external_id can create ill-formed URIs, so we create websafe URIs
                # since labels do not have internal ids, we use the external_id as the id
                triples.append(
                    (id_, namespace.label, namespace[f"Label_{create_sha256_hash(label.dump()['externalId'])}"])
                )

        if asset.metadata:
            for key, value in asset.metadata.items():
                if value:
                    triples.append((id_, namespace[key], Literal(string_to_ideal_type(value))))

        # Create connections:
        if asset.parent_id:
            triples.append((id_, namespace.parent, namespace[f"Asset_{asset.parent_id}"]))

        if asset.root_id:
            triples.append((id_, namespace.root, namespace[f"Asset_{asset.root_id}"]))

        if asset.data_set_id:
            triples.append((id_, namespace.dataset, namespace[f"Dataset_{asset.data_set_id}"]))

        return triples



import json
import logging
import os
import threading
import time
import traceback
from pathlib import Path
from threading import Event

import yaml
from cognite.client import ClientConfig, CogniteClient
from prometheus_client import Gauge

from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.config import Config
from cognite.neat.utils.utils import retry_decorator
from cognite.neat.workflows import cdf_store, utils
from cognite.neat.workflows._exceptions import ConfigurationNotSet, InvalidStepOutputException
from cognite.neat.workflows.cdf_store import CdfStore
from cognite.neat.workflows.model import (
    FlowMessage,
    StepExecutionStatus,
    StepType,
    WorkflowConfigItem,
    WorkflowConfigs,
    WorkflowDefinition,
    WorkflowFullStateReport,
    WorkflowStartException,
    WorkflowState,
    WorkflowStepDefinition,
    WorkflowStepEvent,
    WorkflowSystemComponent,
)
from cognite.neat.workflows.steps.step_model import DataContract
from cognite.neat.workflows.steps_registry import StepsRegistry
from cognite.neat.workflows.tasks import WorkflowTaskBuilder

summary_metrics = Gauge("neat_workflow_summary_metrics", "Workflow execution summary metrics", ["wf_name", "name"])
steps_metrics = Gauge("neat_workflow_steps_metrics", "Workflow step level metrics", ["wf_name", "step_name", "name"])
timing_metrics = Gauge("neat_workflow_timing_metrics", "Workflow timing metrics", ["wf_name", "component", "name"])


class BaseWorkflow:
    def __init__(
        self,
        name: str,
        client: CogniteClient,
        steps_registry: StepsRegistry,
        workflow_steps: list[WorkflowStepDefinition] | None = None,
        default_dataset_id: int | None = None,
    ):
        self.name = name
        self.module_name = self.__class__.__module__
        self.cdf_client = client
        self.cdf_client_config: ClientConfig = client.config
        self.default_dataset_id = default_dataset_id
        self.state = WorkflowState.CREATED
        self.instance_id = utils.generate_run_id()
        self.run_id = ""
        self.last_error = ""
        self.elapsed_time: float = 0.0
        self.start_time: float | None = None
        self.end_time: float | None = None
        self.execution_log: list[WorkflowStepEvent] = []
        self.workflow_steps: list[WorkflowStepDefinition] = workflow_steps or []
        self.workflow_system_components: list[WorkflowSystemComponent] = []
        self.configs: list[WorkflowConfigItem] = []
        self.flow_message: FlowMessage | None = None
        self.task_builder: WorkflowTaskBuilder | None = None
        self.rules_storage_path: Path | None = None
        self.data_store_path: Path | None = None
        self.user_steps_path: Path | None = None  # path to user defined steps
        self.cdf_store = (
            cdf_store.CdfStore(self.cdf_client, data_set_id=self.default_dataset_id)
            if self.default_dataset_id
            else None
        )
        self.metrics = NeatMetricsCollector(self.name, self.cdf_client)
        self.resume_event = Event()
        self.is_ephemeral = False  # if True, workflow will be deleted after completion
        self.auto_workflow_cleanup = False
        self.step_classes = None
        self.data: dict[str, DataContract | FlowMessage | CdfStore | CogniteClient | None] = {}
        self.steps_registry: StepsRegistry = steps_registry

    def start(self, sync=False, is_ephemeral=False, **kwargs) -> FlowMessage | None:
        """Starts workflow execution.sync=True will block until workflow is completed and
        return last workflow flow message, sync=False will start workflow in a separate thread and return None"""
        if self.state not in [WorkflowState.CREATED, WorkflowState.COMPLETED, WorkflowState.FAILED]:
            logging.error(f"Workflow {self.name} is already running")
            return None
        self.data["StartFlowMessage"] = kwargs.get("flow_message", None)
        self.data["CdfStore"] = self.cdf_store
        self.data["CogniteClient"] = self.cdf_client
        self.state = WorkflowState.RUNNING
        self.start_time = time.time()
        self.end_time = None
        self.run_id = utils.generate_run_id()
        self.is_ephemeral = is_ephemeral
        self.execution_log = []

        if sync:
            return self._run_workflow(**kwargs)

        self.thread = threading.Thread(target=self._run_workflow, kwargs=kwargs)
        self.thread.start()
        return None

    def _run_workflow(self, **kwargs) -> FlowMessage | None:
        """Run workflow and return last workflow flow message"""
        summary_metrics.labels(wf_name=self.name, name="steps_count").set(len(self.workflow_steps))
        logging.info(f"Starting workflow {self.name}")
        if flow_message := kwargs.get("flow_message"):
            self.flow_message = flow_message
            self.data["FlowMessage"] = flow_message
        start_time = time.perf_counter()
        self.report_workflow_execution()
        try:
            start_step_id = kwargs.get("start_step_id")
            logging.info(f"  starting workflow from step {start_step_id}")

            self.run_workflow_steps(start_step_id=start_step_id)
            if self.state == WorkflowState.RUNNING:
                self.state = WorkflowState.COMPLETED
            summary_metrics.labels(wf_name=self.name, name="wf_completed_counter").inc()
        except Exception:
            trace = traceback.format_exc()
            self.last_error = str(trace)
            self.state = WorkflowState.FAILED
            logging.error(f"Workflow failed with error {trace}")
            summary_metrics.labels(wf_name=self.name, name="wf_failed_counter").inc()

        self.elapsed_time = time.perf_counter() - start_time
        self.end_time = time.time()
        timing_metrics.labels(wf_name=self.name, component="workflow", name="workflow").set(self.elapsed_time)
        logging.info(f"Workflow completed in {self.elapsed_time} seconds")
        self.report_workflow_execution()
        if self.auto_workflow_cleanup:
            self.cleanup_workflow_context()
        return self.flow_message

    def cleanup_workflow_context(self):
        """
        Cleans up the workflow context by closing and deleting graph stores and other structure.
        It's a for of garbage collection to avoid memory leaks or other issues related to open handlers
        or allocated resources

        """
        if "SolutionGraph" in self.data:
            self.data["SolutionGraph"].graph.close()
        if "SourceGraph" in self.data:
            self.data["SourceGraph"].graph.close()
        self.data.clear()

    def get_transition_step(self, transitions: list[str] | None) -> list[WorkflowStepDefinition]:
        return [stp for stp in self.workflow_steps if stp.id in transitions and stp.enabled] if transitions else []

    def run_workflow_steps(self, start_step_id: str | None = None) -> str:
        if not start_step_id:
            trigger_steps = list(filter(lambda x: x.trigger, self.workflow_steps))
        else:
            trigger_steps = list(filter(lambda x: x.id == start_step_id, self.workflow_steps))

        if not trigger_steps:
            logging.error(f"Workflow {self.name} has no trigger steps or start step {start_step_id} not found")
            return "Workflow has no trigger steps"

        self.execution_log.append(
            WorkflowStepEvent(
                id=trigger_steps[0].id,
                state=StepExecutionStatus.SUCCESS,
                elapsed_time=0,
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                data=self.flow_message.payload if self.flow_message else None,
            )
        )

        step: WorkflowStepDefinition = trigger_steps[0]
        transition_steps = self.get_transition_step(step.transition_to)

        if len(transition_steps) == 0:
            logging.error(f"Workflow {self.name} has no transition steps from step {step.id}")
            return "Workflow has no transition steps"

        step = transition_steps[0]

        for _ in range(
            1000
        ):  # Max 1000 steps in a workflow is a reasonable limit and protection against infinite loops
            new_flow_msg = None
            if step.enabled:
                new_flow_msg = self.run_step(step)
            else:
                logging.info(f"Skipping step workflow step {step.id}")
                self.execution_log.append(
                    WorkflowStepEvent(
                        id=step.id,
                        system_component_id=step.system_component_id,
                        state=StepExecutionStatus.SKIPPED,
                        elapsed_time=0,
                        timestamp=utils.get_iso8601_timestamp_now_unaware(),
                        error="",
                    )
                )

            steps = self.get_transition_step(step.transition_to)

            if new_flow_msg:
                self.flow_message = new_flow_msg
                # If the step returned a new flow message, use it to mutate execution flow
                if new_flow_msg.next_step_ids:
                    steps = self.get_transition_step(new_flow_msg.next_step_ids)

            if len(steps) == 0:
                break
            step = steps[0]
            logging.debug(f"Transitioning to step {step.id}")

        return "ok"

    def configure(self, config: Config):
        raise NotImplementedError()

    def copy(self) -> "BaseWorkflow":
        """Create a copy of the workflow"""
        new_instance = self.__class__(self.name, self.cdf_client, self.steps_registry)
        new_instance.workflow_steps = self.workflow_steps
        new_instance.configs = self.configs
        new_instance.set_task_builder(self.task_builder)
        new_instance.set_default_dataset_id(self.default_dataset_id)
        new_instance.set_storage_path("transformation_rules", self.rules_storage_path)
        if self.data_store_path:
            new_instance.set_storage_path("data_store", self.data_store_path)
        return new_instance

    def run_step(self, step: WorkflowStepDefinition) -> FlowMessage | None:
        step_name = step.id
        system_component_id = step.system_component_id

        steps_metrics.labels(wf_name=self.name, step_name=step_name, name="step_started_counter").inc()
        flow_message = self.flow_message
        if self.state != WorkflowState.RUNNING:
            logging.error(f"Workflow {self.name} is not running , step {step_name} is skipped")
            return None

        logging.info(f"Running step {step_name}")
        self.current_step = step_name
        start_time = time.perf_counter()
        stop_time = start_time
        step_execution_status = StepExecutionStatus.STARTED
        self.execution_log.append(
            WorkflowStepEvent(
                id=step_name,
                system_component_id=system_component_id,
                state=step_execution_status,
                elapsed_time=0,
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                error="",
            )
        )
        output_text = ""
        error_text = ""
        new_flow_message = None
        try:
            if step.stype == StepType.PYSTEP:
                # Most likely will be discontinued in the future in favor of std steps
                if step.method and hasattr(self, step.method):
                    method = getattr(self, step.method)
                else:
                    method_name = f"step_{step.id}"
                    if hasattr(self, method_name):
                        method = getattr(self, method_name)
                    else:
                        logging.error(f"Workflow step {step.id} has no method {method_name}")
                        raise Exception(f"Workflow step {step.id} has no method {method_name}")

                @retry_decorator(
                    max_retries=step.max_retries,
                    retry_delay=step.retry_delay,
                    component_name=f"wf step runner , step.id = {step.id}",
                )
                def method_runner():
                    return method(flow_message)

                new_flow_message = method_runner()
            elif step.stype == StepType.STD_STEP:
                if self.steps_registry is None:
                    raise Exception(
                        f"Workflow step {step.id} can't be executed.Step registry is not configured or \
                          not set as parameter in BaseWorkflow constructor"
                    )

                @retry_decorator(
                    max_retries=step.max_retries,
                    retry_delay=step.retry_delay,
                    component_name=f"wf step runner , step.id = {step.id}",
                )
                def method_runner():
                    return self.steps_registry.run_step(
                        step.method,
                        self.data,
                        metrics=self.metrics,
                        workflow_configs=self.get_configs(),
                        step_configs=step.configs,
                        workflow_id=self.name,
                        workflow_run_id=self.run_id,
                        step_complex_configs=step.complex_configs,
                    )

                output = method_runner()
                if output is not None:
                    outputs = output if isinstance(output, tuple) else (output,)
                    for out_obj in outputs:
                        if isinstance(out_obj, FlowMessage):
                            new_flow_message = out_obj
                        elif isinstance(out_obj, DataContract):
                            self.data[type(out_obj).__name__] = out_obj
                        else:
                            raise InvalidStepOutputException(step_type=type(out_obj).__name__)

            elif step.stype == StepType.START_WORKFLOW_TASK_STEP:
                if self.task_builder:
                    sync_str = step.params.get("sync", "false")
                    sync = sync_str.lower() == "true" or sync_str == "1"
                    start_status = self.task_builder.start_workflow_task(
                        workflow_name=step.params.get("workflow_name", ""), sync=sync, flow_message=self.flow_message
                    )
                    if start_status.is_success and start_status.workflow_instance.state == WorkflowState.COMPLETED:
                        # It only works with sync = true
                        new_flow_message = start_status.workflow_instance.flow_message
                    else:
                        logging.error(f"Workflow step {step.id} failed to start workflow task")
                        if start_status.is_success:
                            raise WorkflowStartException(start_status.workflow_instance.last_error)
                        else:
                            raise WorkflowStartException(start_status.status_text)

                else:
                    logging.error(f"Workflow step {step.id} has no task builder")
                    raise Exception(f"Workflow step {step.id} has no task builder")
            elif step.stype == StepType.WAIT_FOR_EVENT:
                # Pause workflow execution until event is received
                if self.state != WorkflowState.RUNNING:
                    logging.error(f"Workflow {self.name} is not running , step {step_name} is skipped")
                    raise Exception(f"Workflow {self.name} is not running , step {step_name} is skipped")
                self.state = WorkflowState.RUNNING_WAITING
                timeout = 3000000.0
                if step.params.get("wait_timeout"):
                    timeout = float(step.params["wait_timeout"])
                # reporting workflow execution before waiting for event
                logging.info(f"Workflow {self.name} is waiting for event")
                self.resume_event.wait(timeout=timeout)
                logging.info(f"Workflow {self.name} resumed after event")
                self.state = WorkflowState.RUNNING
                self.resume_event.clear()
            else:
                logging.error(f"Workflow step {step.id} has unsupported step type {step.stype}")

            stop_time = time.perf_counter()
            elapsed_time = stop_time - start_time
            logging.info(f"Step {step_name} completed in {elapsed_time} seconds with ")

            if new_flow_message:
                logging.info(f"Step {self.name} completed with status {new_flow_message.step_execution_status}")
                error_text = new_flow_message.error_text
                output_text = new_flow_message.output_text
                if new_flow_message.step_execution_status == StepExecutionStatus.ABORT_AND_FAIL:
                    new_flow_message.step_execution_status = StepExecutionStatus.FAILED
                    self.state = WorkflowState.FAILED

                step_execution_status = (
                    StepExecutionStatus.SUCCESS
                    if new_flow_message.step_execution_status == StepExecutionStatus.UNKNOWN
                    else new_flow_message.step_execution_status
                )
            else:
                step_execution_status = StepExecutionStatus.SUCCESS
            steps_metrics.labels(wf_name=self.name, step_name=step_name, name="completed_counter").inc()
            timing_metrics.labels(wf_name=self.name, component="step", name=step_name).set(elapsed_time)
        except Exception:
            self.state = WorkflowState.FAILED
            step_execution_status = StepExecutionStatus.FAILED
            trace = traceback.format_exc()
            elapsed_time = stop_time - start_time
            logging.error(f"Step {step_name} failed with error : {trace}")
            error_text = str(trace)
            self.last_error = error_text
            traceback.print_exc()
            steps_metrics.labels(wf_name=self.name, step_name=step_name, name="failed_counter").inc()

        self.execution_log.append(
            WorkflowStepEvent(
                id=step_name,
                system_component_id=system_component_id,
                state=step_execution_status,
                elapsed_time=round(elapsed_time, 3),
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                error=error_text,
                output_text=output_text,
                data=new_flow_message.payload if new_flow_message else None,
            )
        )
        if new_flow_message:
            self.flow_message = new_flow_message

        self.report_step_execution()
        return new_flow_message

    def resume_workflow(self, flow_message: FlowMessage, step_id: str):
        if step_id and self.current_step != step_id:
            logging.error(f"Workflow {self.name} is not in step {step_id} , resume is skipped")
            return
        self.flow_message = flow_message
        self.execution_log.append(
            WorkflowStepEvent(
                id=step_id,
                state=StepExecutionStatus.SUCCESS,
                elapsed_time=0,
                timestamp=utils.get_iso8601_timestamp_now_unaware(),
                data=self.flow_message.payload,
            )
        )
        self.resume_event.set()

    def report_step_execution(self):
        pass

    def report_workflow_execution(self):
        reporting_type = "all_enabled"
        if config_item := self.get_config_item("system.execution_reporting_type"):
            reporting_type = config_item.value
        if reporting_type == "all_disabled":
            return
        if self.cdf_store:
            logging.debug("Reporting workflow execution to CDF")
            try:
                self.cdf_store.report_workflow_execution_to_cdf(self.get_state())
            except Exception:
                logging.error("Failed to report workflow execution to CDF")
                traceback.print_exc()

    def get_state(self):
        return WorkflowFullStateReport(
            workflow_name=self.name,
            state=self.state,
            run_id=self.run_id,
            start_time=self.start_time,
            end_time=self.end_time,
            elapsed_time=round(self.elapsed_time, 3),
            last_error=self.last_error,
            execution_log=self.execution_log,
        )

    def set_cognite_client(self, client: CogniteClient):
        self.cdf_client = client

    def get_workflow_definition(self):
        return WorkflowDefinition(
            name=self.name,
            steps=self.workflow_steps,
            system_components=self.workflow_system_components,
            configs=self.configs,
        )

    def add_step(self, step: WorkflowStepDefinition):
        self.workflow_steps.append(step)

    def enable_step(self, step_id: str, enabled: bool = True):
        """Enable or disable step in workflow. Primarily used for tests"""
        for step in self.workflow_steps:
            if step.id == step_id:
                step.enabled = enabled

    def add_system_component(self, system_components: WorkflowSystemComponent):
        self.workflow_system_components.append(system_components)

    def serialize_workflow(self, output_format: str = "json", custom_implementation_module: str | None = None) -> str:
        workflow_definitions = WorkflowDefinition(
            name=self.name,
            steps=self.workflow_steps,
            system_components=self.workflow_system_components,
            configs=self.configs,
            implementation_module=custom_implementation_module,
        )
        if output_format == "json":
            return json.dumps(workflow_definitions.model_dump(), indent=4)
        elif output_format == "yaml":
            return yaml.dump(workflow_definitions.model_dump(), indent=4)
        else:
            raise NotImplementedError(f"Output format {output_format} is not supported.")

    @classmethod
    def deserialize_definition(cls, json_string: str, output_format: str = "json") -> WorkflowDefinition:
        if output_format == "json":
            workflow_definitions = WorkflowDefinition.model_validate(json.loads(json_string))
        elif output_format == "yaml":
            workflow_definitions = WorkflowDefinition.model_validate(yaml.load(json_string, Loader=yaml.Loader))
        else:
            raise NotImplementedError(f"Output format {output_format} is not supported.")
        return workflow_definitions

    def set_definition(self, workflow_definition: WorkflowDefinition):
        self.workflow_steps = workflow_definition.steps
        self.workflow_system_components = workflow_definition.system_components
        self.configs = workflow_definition.configs

    def set_storage_path(self, storage_type: str, storage_path: str | Path | None) -> None:
        if storage_path is None:
            return None
        if storage_type == "transformation_rules":
            self.rules_storage_path = Path(storage_path)
            if self.default_dataset_id is None:
                raise ConfigurationNotSet("default_dataset_id")
            self.cdf_store = cdf_store.CdfStore(
                self.cdf_client, data_set_id=self.default_dataset_id, rules_storage_path=self.rules_storage_path
            )
        elif storage_type == "data_store":
            self.data_store_path = Path(storage_path)
            self.user_steps_path = Path(storage_path, "steps")

    def set_task_builder(self, task_builder: WorkflowTaskBuilder | None):
        self.task_builder = task_builder

    def get_config_item(self, config_name: str) -> WorkflowConfigItem | None:
        return next((item for item in self.configs if item.name == config_name), None)

    def get_config_group_values_by_name(
        self, group_name: str, remove_group_prefix: bool = True
    ) -> dict[str, str | None]:
        return {
            (item.name.removeprefix(item.group) if remove_group_prefix else item.name): item.value
            for item in self.configs
            if item.group == group_name
        }

    def get_config_item_value(self, config_name: str, default_value=None) -> str | None:
        return config.value if (config := self.get_config_item(config_name)) else default_value

    def set_default_dataset_id(self, default_dataset_id: int | None):
        self.default_dataset_id = default_dataset_id
        self.cdf_store = (
            cdf_store.CdfStore(self.cdf_client, data_set_id=self.default_dataset_id)
            if self.default_dataset_id
            else None
        )

    def set_cdf_client_config(self, cdf_client_config: ClientConfig):
        """Set the CogniteClient configuration to be used by the workflow to create new CogniteClient instances."""
        self.cdf_client_config = cdf_client_config

    def get_new_cdf_client(self):
        """Get a new CogniteClient instance with the same configuration as the one used by the workflow .
          Should be used from workflow steps to avoid sharing the same client instance between steps or
          reset reference to old client instance.
        Returns: CogniteClient
        """
        return CogniteClient(self.cdf_client_config)

    def get_step_by_id(self, step_id: str) -> WorkflowStepDefinition | None:
        return next((step for step in self.workflow_steps if step.id == step_id), None)

    def get_trigger_step(self, step_id: str | None = None) -> WorkflowStepDefinition | None:
        if step_id:
            return next((step for step in self.workflow_steps if step.id == step_id and step.enabled), None)
        else:
            return next((step for step in self.workflow_steps if step.trigger and step.enabled), None)

    def get_context(self) -> dict[str, DataContract | FlowMessage | CdfStore | CogniteClient | None]:
        return self.data

    def get_configs(self) -> WorkflowConfigs:
        return WorkflowConfigs(configs=self.configs)

    def get_list_of_workflow_artifacts(self) -> list[Path]:
        # create a list of all files under the data store path

        file_list: list[Path] = []
        if self.data_store_path is None:
            raise ConfigurationNotSet("data_store_path")
        workflow_data_path = Path(self.data_store_path) / "workflows" / self.name
        try:
            for root, _dirs, files in os.walk(workflow_data_path):
                for file in files:
                    full_path = Path(root) / file
                    file_list.append(Path(full_path).relative_to(workflow_data_path))
            return file_list
        except OSError as e:
            # Handle exceptions like directory not found, permission denied, etc.
            logging.error(f"Error while listing workflow artifacts for {self.name} : {e}")
            return []



import re
from datetime import datetime
from pathlib import Path
from typing import Literal, cast, overload

from rdflib import Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral

import cognite.neat.rules.issues as issues
from cognite.neat.constants import DEFAULT_NAMESPACE, PREFIXES
from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.rules.importers._base import BaseImporter, Rules, _handle_issues
from cognite.neat.rules.issues import IssueList
from cognite.neat.rules.models import InformationRules, RoleTypes
from cognite.neat.rules.models._base import MatchType
from cognite.neat.rules.models.information import (
    InformationMetadata,
    InformationRulesInput,
)
from cognite.neat.utils.utils import get_namespace, remove_namespace, uri_to_short_form

ORDERED_CLASSES_QUERY = """SELECT ?class (count(?s) as ?instances )
                           WHERE { ?s a ?class . }
                           group by ?class order by DESC(?instances)"""

INSTANCES_OF_CLASS_QUERY = """SELECT ?s WHERE { ?s a <class> . }"""

INSTANCE_PROPERTIES_DEFINITION = """SELECT ?property (count(?property) as ?occurrence) ?dataType ?objectType
                                    WHERE {<instance_id> ?property ?value .
                                           BIND(datatype(?value) AS ?dataType)
                                           OPTIONAL {?value rdf:type ?objectType .}}
                                    GROUP BY ?property ?dataType ?objectType"""


class InferenceImporter(BaseImporter):
    """Infers rules from a triple store.

    Rules inference through analysis of knowledge graph provided in various formats.
    Use the factory methods to create an triples store from sources such as
    RDF files, JSON files, YAML files, XML files, or directly from a graph store.

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        max_number_of_instance: Maximum number of instances to be used in inference
    """

    def __init__(self, issue_list: IssueList, graph: Graph, max_number_of_instance: int = -1):
        self.issue_list = issue_list
        self.graph = graph
        self.max_number_of_instance = max_number_of_instance

    @classmethod
    def from_graph_store(cls, store: NeatGraphStore, max_number_of_instance: int = -1):
        issue_list = IssueList(title="Inferred from graph store")

        return cls(issue_list, store.graph, max_number_of_instance=max_number_of_instance)

    @classmethod
    def from_rdf_file(cls, filepath: Path, max_number_of_instance: int = -1):
        issue_list = IssueList(title=f"'{filepath.name}'")

        graph = Graph()
        try:
            graph.parse(filepath)
        except Exception:
            issue_list.append(issues.fileread.FileReadError(filepath))

        return cls(issue_list, graph, max_number_of_instance=max_number_of_instance)

    @classmethod
    def from_json_file(cls, filepath: Path, max_number_of_instance: int = -1):
        raise NotImplementedError("JSON file format is not supported yet.")

    @classmethod
    def from_yaml_file(cls, filepath: Path, max_number_of_instance: int = -1):
        raise NotImplementedError("YAML file format is not supported yet.")

    @classmethod
    def from_xml_file(cls, filepath: Path, max_number_of_instance: int = -1):
        raise NotImplementedError("JSON file format is not supported yet.")

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules: ...

    @overload
    def to_rules(
        self,
        errors: Literal["continue"] = "continue",
        role: RoleTypes | None = None,
    ) -> tuple[Rules | None, IssueList]: ...

    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        """
        Creates `Rules` object from the data for target role.
        """

        if self.issue_list.has_errors:
            # In case there were errors during the import, the to_rules method will return None
            return self._return_or_raise(self.issue_list, errors)

        rules_dict = self._to_rules_components()

        # adding additional prefix
        rules_dict["prefixes"][rules_dict["metadata"]["prefix"]] = rules_dict["metadata"]["namespace"]

        with _handle_issues(self.issue_list) as future:
            rules: InformationRules
            rules = InformationRulesInput.load(rules_dict).as_rules()

        if future.result == "failure" or self.issue_list.has_errors:
            return self._return_or_raise(self.issue_list, errors)

        return self._to_output(
            rules,
            self.issue_list,
            errors=errors,
            role=role,
        )

    def _to_rules_components(
        self,
    ) -> dict:
        """Convert RDF graph to dictionary defining data model and prefixes of the graph

        Args:
            graph: RDF graph to be converted to TransformationRules object
            max_number_of_instance: Max number of instances to be considered for each class

        Returns:
            Tuple of data model and prefixes of the graph
        """
        classes: dict[str, dict] = {}
        properties: dict[str, dict] = {}
        prefixes: dict[str, Namespace] = PREFIXES

        # Infers all the classes in the graph
        for class_uri, no_instances in self.graph.query(ORDERED_CLASSES_QUERY):  # type: ignore[misc]
            self._add_uri_namespace_to_prefixes(cast(URIRef, class_uri), prefixes)

            if (class_id := remove_namespace(class_uri)) in classes:
                # handles cases when class id is already present in classes
                class_id = f"{class_id}_{len(classes)+1}"

            classes[class_id] = {
                "class_": class_id,
                "reference": class_uri,
                "match_type": MatchType.exact,
                "comment": f"Inferred from knowledge graph, where this class has <{no_instances}> instances",
            }

        # Infers all the properties of the class
        for class_id, class_definition in classes.items():
            for (instance,) in self.graph.query(  # type: ignore[misc]
                INSTANCES_OF_CLASS_QUERY.replace("class", class_definition["reference"])
                if self.max_number_of_instance < 0
                else INSTANCES_OF_CLASS_QUERY.replace("class", class_definition["reference"])
                + f" LIMIT {self.max_number_of_instance}"
            ):
                for property_uri, occurrence, data_type_uri, object_type_uri in self.graph.query(  # type: ignore[misc]
                    INSTANCE_PROPERTIES_DEFINITION.replace("instance_id", instance)
                ):  # type: ignore[misc]
                    property_id = remove_namespace(property_uri)
                    self._add_uri_namespace_to_prefixes(cast(URIRef, property_uri), prefixes)
                    value_type_uri = data_type_uri if data_type_uri else object_type_uri

                    # this is to skip rdf:type property
                    if not value_type_uri:
                        continue

                    self._add_uri_namespace_to_prefixes(cast(URIRef, value_type_uri), prefixes)
                    value_type_id = remove_namespace(value_type_uri)
                    id_ = f"{class_id}:{property_id}"

                    definition = {
                        "class_": class_id,
                        "property_": property_id,
                        "max_count": cast(RdfLiteral, occurrence).value,
                        "value_type": value_type_id,
                        "reference": property_uri,
                        "transformation": (
                            f"{uri_to_short_form(class_definition['reference'], prefixes)}"
                            f"({uri_to_short_form(cast(URIRef, property_uri), prefixes)})"
                        ),
                        "comment": (
                            f"Class <{class_id}> has property <{property_id}> with "
                            f"value type <{value_type_id}> which occurs <1> times in the graph"
                        ),
                    }

                    # USE CASE 1: If property is not present in properties
                    if id_ not in properties:
                        properties[id_] = definition

                    # USE CASE 2: first time redefinition, value type change to multi
                    elif id_ in properties and definition["value_type"] not in properties[id_]["value_type"]:
                        properties[id_]["value_type"] = properties[id_]["value_type"] + " | " + definition["value_type"]
                        properties[id_]["comment"] = (
                            properties[id_]["comment"] + ", with" + definition["comment"].split("with")[1]
                        )

                    # USE CASE 3: existing but max count is different
                    elif (
                        id_ in properties
                        and definition["value_type"] in properties[id_]["value_type"]
                        and not (properties[id_]["max_count"] == definition["max_count"])
                    ):
                        properties[id_]["max_count"] = max(properties[id_]["max_count"], definition["max_count"])

                        properties[id_]["comment"] = self._update_value_type_occurrence_in_comment(
                            definition["value_type"], properties[id_]["comment"]
                        )

                    # USE CASE 4: Just update the comment with occurrence
                    else:
                        properties[id_]["comment"] = self._update_value_type_occurrence_in_comment(
                            definition["value_type"], properties[id_]["comment"]
                        )

        return {
            "metadata": self._default_metadata().model_dump(),
            "classes": list(classes.values()),
            "properties": list(properties.values()),
            "prefixes": prefixes,
        }

    @classmethod
    def _add_uri_namespace_to_prefixes(cls, URI: URIRef, prefixes: dict[str, Namespace]):
        """Add URI to prefixes dict if not already present

        Args:
            URI: URI from which namespace is being extracted
            prefixes: Dict of prefixes and namespaces
        """
        if Namespace(get_namespace(URI)) not in prefixes.values():
            prefixes[f"prefix-{len(prefixes)+1}"] = Namespace(get_namespace(URI))

    @classmethod
    def _default_metadata(cls):
        return InformationMetadata(
            name="Inferred Model",
            creator="NEAT",
            version="inferred",
            created=datetime.now(),
            updated=datetime.now(),
            description="Inferred model from knowledge graph",
            prefix="neat",
            namespace=DEFAULT_NAMESPACE,
        )

    @classmethod
    def _update_value_type_occurrence_in_comment(cls, value_type: str, comment: str) -> str:
        occurrence = cls._read_value_type_occurrence_from_comment(value_type, comment)
        return comment.replace(
            f"with value type <{value_type}> which occurs <{occurrence}> times in the graph",
            f"with value type <{value_type}> which occurs <{occurrence+1}> times in the graph",
        )

    @classmethod
    def _read_value_type_occurrence_from_comment(cls, value_type: str, comment: str) -> int:
        return int(
            cast(
                re.Match, re.search(rf"with value type <{value_type}> which occurs <(\d+)> times in the graph", comment)
            ).group(1)
        )



import itertools
import json
from collections import defaultdict
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.capabilities import Capability, DataModelInstancesAcl
from cognite.client.data_classes.data_modeling import ViewId
from cognite.client.data_classes.data_modeling.ids import InstanceId
from cognite.client.data_classes.data_modeling.views import SingleEdgeConnection
from cognite.client.exceptions import CogniteAPIError
from pydantic import ValidationInfo, create_model, field_validator
from pydantic.main import Model

from cognite.neat.graph._tracking import LogTracker, Tracker
from cognite.neat.graph.issues import loader as loader_issues
from cognite.neat.graph.stores import NeatGraphStore
from cognite.neat.issues import NeatIssue, NeatIssueList
from cognite.neat.rules.models import DMSRules
from cognite.neat.rules.models.data_types import _DATA_TYPE_BY_DMS_TYPE
from cognite.neat.utils.upload import UploadResult
from cognite.neat.utils.utils import create_sha256_hash

from ._base import CDFLoader


class DMSLoader(CDFLoader[dm.InstanceApply]):
    """Load data from Cognite Data Fusions Data Modeling Service (DMS) into Neat.

    Args:
        graph_store (NeatGraphStore): The graph store to load the data into.
        data_model (dm.DataModel[dm.View] | None): The data model to load.
        instance_space (str): The instance space to load the data into.
        class_by_view_id (dict[ViewId, str] | None): A mapping from view id to class name. Defaults to None.
        creat_issues (Sequence[NeatIssue] | None): A list of issues that occurred during reading. Defaults to None.
        tracker (type[Tracker] | None): The tracker to use. Defaults to None.
    """

    def __init__(
        self,
        graph_store: NeatGraphStore,
        data_model: dm.DataModel[dm.View] | None,
        instance_space: str,
        class_by_view_id: dict[ViewId, str] | None = None,
        create_issues: Sequence[NeatIssue] | None = None,
        tracker: type[Tracker] | None = None,
    ):
        super().__init__(graph_store)
        self.data_model = data_model
        self.instance_space = instance_space
        self.class_by_view_id = class_by_view_id or {}
        self._issues = NeatIssueList[NeatIssue](create_issues or [])
        self._tracker: type[Tracker] = tracker or LogTracker

    @classmethod
    def from_data_model_id(
        cls,
        client: CogniteClient,
        data_model_id: dm.DataModelId,
        graph_store: NeatGraphStore,
        instance_space: str,
    ) -> "DMSLoader":
        issues: list[NeatIssue] = []
        data_model: dm.DataModel[dm.View] | None = None
        try:
            data_model = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True).latest_version()
        except Exception as e:
            issues.append(loader_issues.MissingDataModelError(identifier=repr(data_model_id), reason=str(e)))

        return cls(graph_store, data_model, instance_space, {}, issues)

    @classmethod
    def from_rules(cls, rules: DMSRules, graph_store: NeatGraphStore, instance_space: str) -> "DMSLoader":
        issues: list[NeatIssue] = []
        data_model: dm.DataModel[dm.View] | None = None
        try:
            data_model = rules.as_schema().as_read_model()
        except Exception as e:
            issues.append(
                loader_issues.FailedConvertError(
                    identifier=rules.metadata.as_identifier(), target_format="read DMS model", reason=str(e)
                )
            )
        return cls(graph_store, data_model, instance_space, {}, issues)

    def _load(self, stop_on_exception: bool = False) -> Iterable[dm.InstanceApply | NeatIssue]:
        if self._issues.has_errors and stop_on_exception:
            raise self._issues.as_exception()
        elif self._issues.has_errors:
            yield from self._issues
            return
        if not self.data_model:
            # There should already be an error in this case.
            return
        view_ids = [repr(v.as_id()) for v in self.data_model.views]
        tracker = self._tracker(type(self).__name__, view_ids, "views")
        for view in self.data_model.views:
            view_id = view.as_id()
            tracker.start(repr(view_id))
            pydantic_cls, edge_by_properties, issues = self._create_validation_classes(view)  # type: ignore[var-annotated]
            yield from issues
            tracker.issue(issues)
            class_name = self.class_by_view_id.get(view.as_id(), view.external_id)
            triples = self.graph_store.read(class_name)
            for identifier, properties in _triples2dictionary(triples).items():
                try:
                    yield self._create_node(identifier, properties, pydantic_cls, view_id)
                except ValueError as e:
                    error = loader_issues.InvalidInstanceError(type_="node", identifier=identifier, reason=str(e))
                    tracker.issue(error)
                    if stop_on_exception:
                        raise error.as_exception() from e
                    yield error
                yield from self._create_edges(identifier, properties, edge_by_properties, tracker)
            tracker.finish(repr(view_id))

    def write_to_file(self, filepath: Path) -> None:
        if filepath.suffix not in [".json", ".yaml", ".yml"]:
            raise ValueError(f"File format {filepath.suffix} is not supported")
        dumped: dict[str, list] = {"nodes": [], "edges": [], "issues": []}
        for item in self.load(stop_on_exception=False):
            key = {
                dm.NodeApply: "nodes",
                dm.EdgeApply: "edges",
                NeatIssue: "issues",
            }.get(type(item))
            if key is None:
                # This should never happen, and is a bug in neat
                raise ValueError(f"Item {item} is not supported. This is a bug in neat please report it.")
            dumped[key].append(item.dump())
        with filepath.open("w", encoding=self._encoding, newline=self._new_line) as f:
            if filepath.suffix == ".json":
                json.dump(dumped, f, indent=2)
            else:
                yaml.safe_dump(dumped, f, sort_keys=False)

    def _create_validation_classes(
        self, view: dm.View
    ) -> tuple[type[Model], dict[str, dm.EdgeConnection], NeatIssueList]:
        issues = NeatIssueList[NeatIssue]()
        field_definitions: dict[str, tuple[type, Any]] = {}
        edge_by_property: dict[str, dm.EdgeConnection] = {}
        direct_relation_by_property: dict[str, dm.DirectRelation] = {}
        for prop_name, prop in view.properties.items():
            if isinstance(prop, dm.EdgeConnection):
                edge_by_property[prop_name] = prop
            if isinstance(prop, dm.MappedProperty):
                if isinstance(prop.type, dm.DirectRelation):
                    direct_relation_by_property[prop_name] = prop.type
                    python_type: Any = dict
                else:
                    data_type = _DATA_TYPE_BY_DMS_TYPE.get(prop.type._type)
                    if not data_type:
                        issues.append(
                            loader_issues.InvalidClassWarning(
                                class_name=repr(view.as_id()),
                                reason=f"Unknown data type for property {prop_name}: {prop.type._type}",
                            )
                        )
                        continue
                    python_type = data_type.python
                if prop.type.is_list:
                    python_type = list[python_type]
                default_value: Any = prop.default_value
                if prop.nullable:
                    python_type = python_type | None
                else:
                    default_value = ...

                field_definitions[prop_name] = (python_type, default_value)

        def parse_list(cls, value: Any, info: ValidationInfo) -> list[str]:
            if isinstance(value, list) and cls.model_fields[info.field_name].annotation is not list:
                if len(value) == 1:
                    return value[0]
                raise ValueError(f"Got multiple values for {info.field_name}: {value}")
            return value

        validators: dict[str, classmethod] = {"parse_list": field_validator("*", mode="before")(parse_list)}  # type: ignore[dict-item,arg-type]
        if direct_relation_by_property:

            def parse_direct_relation(cls, value: list, info: ValidationInfo) -> dict | list[dict]:
                # We validate above that we only get one value for single direct relations.
                if cls.model_fields[info.field_name].annotation is list:
                    return [{"space": self.instance_space, "externalId": v} for v in value]
                elif value:
                    return {"space": self.instance_space, "externalId": value[0]}
                return {}

            validators["parse_direct_relation"] = field_validator(*direct_relation_by_property.keys(), mode="before")(  # type: ignore[assignment]
                parse_direct_relation  # type: ignore[arg-type]
            )

        pydantic_cls = create_model(view.external_id, __validators__=validators, **field_definitions)  # type: ignore[arg-type, call-overload]
        return pydantic_cls, edge_by_property, issues

    def _create_node(
        self, identifier: str, properties: dict[str, list[str]], pydantic_cls: type[Model], view_id: dm.ViewId
    ) -> dm.InstanceApply:
        created = pydantic_cls.model_validate(properties)

        return dm.NodeApply(
            space=self.instance_space,
            external_id=identifier,
            # type=#RDF type
            sources=[dm.NodeOrEdgeData(source=view_id, properties=dict(created.model_dump().items()))],
        )

    def _create_edges(
        self,
        identifier: str,
        properties: dict[str, list[str]],
        edge_by_properties: dict[str, dm.EdgeConnection],
        tracker: Tracker,
    ) -> Iterable[dm.EdgeApply | NeatIssue]:
        for prop, values in properties.items():
            if prop not in edge_by_properties:
                continue
            edge = edge_by_properties[prop]
            if isinstance(edge, SingleEdgeConnection) and len(values) > 1:
                error = loader_issues.InvalidInstanceError(
                    type_="edge",
                    identifier=identifier,
                    reason=f"Multiple values for single edge {edge}. Expected only one.",
                )
                tracker.issue(error)
                yield error
            for target in values:
                external_id = f"{identifier}.{prop}.{target}"
                yield dm.EdgeApply(
                    space=self.instance_space,
                    external_id=external_id if len(external_id) < 256 else create_sha256_hash(external_id),
                    type=edge.type,
                    start_node=dm.DirectRelationReference(self.instance_space, identifier),
                    end_node=dm.DirectRelationReference(self.instance_space, target),
                )

    def _get_required_capabilities(self) -> list[Capability]:
        return [
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write,
                    DataModelInstancesAcl.Action.Write_Properties,
                    DataModelInstancesAcl.Action.Read,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID([self.instance_space]),
            )
        ]

    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[dm.InstanceApply],
        dry_run: bool,
        read_issues: NeatIssueList,
    ) -> UploadResult:
        result = UploadResult[InstanceId](name=type(self).__name__, issues=read_issues)
        try:
            nodes = [item for item in items if isinstance(item, dm.NodeApply)]
            edges = [item for item in items if isinstance(item, dm.EdgeApply)]
            upserted = client.data_modeling.instances.apply(
                nodes,
                edges,
                auto_create_end_nodes=True,
                auto_create_start_nodes=True,
                skip_on_version_conflict=True,
            )
        except CogniteAPIError as e:
            result.error_messages.append(str(e))
            result.failed_upserted.update(item.as_id() for item in e.failed + e.unknown)
            result.created.update(item.as_id() for item in e.successful)
        else:
            for instance in itertools.chain(upserted.nodes, upserted.edges):
                if instance.was_modified and instance.created_time == instance.last_updated_time:
                    result.created.add(instance.as_id())
                elif instance.was_modified:
                    result.changed.add(instance.as_id())
                else:
                    result.unchanged.add(instance.as_id())
        return result


def _triples2dictionary(
    triples: Iterable[tuple[str, str, str]],
) -> dict[str, dict[str, list[str]]]:
    """Converts list of triples to dictionary"""
    values_by_property_by_identifier: dict[str, dict[str, list[str]]] = defaultdict(lambda: defaultdict(list))
    for id_, property_, value in triples:
        values_by_property_by_identifier[id_][property_].append(value)
    return values_by_property_by_identifier



from abc import ABC
from dataclasses import dataclass
from typing import Any, cast

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

# The Table, TableWrite data classes in the Cognite-SDK lacks the database attribute.
# This is a problem when creating the RawTableLoader that needs the data class to be able to create, update, retrieve
# and delete tables.
# This is a reimplemented version of the Table, TableWrite data classes with the database attribute added.


@dataclass(frozen=True)
class RawTableID:
    table: str
    database: str

    def as_tuple(self) -> tuple[str, str]:
        return self.database, self.table


class RawTableCore(WriteableCogniteResource["RawTableWrite"], ABC):
    """A NoSQL database table to store customer data

    Args:
        name (str | None): Unique name of the table
    """

    def __init__(
        self,
        name: str | None = None,
        database: str | None = None,
    ) -> None:
        self.name = name
        self.database = database

    def as_id(self) -> RawTableID:
        if self.name is None or self.database is None:
            raise ValueError("name and database are required to create a TableID")
        return RawTableID(table=self.name, database=self.database)


class RawTable(RawTableCore):
    """A NoSQL database table to store customer data.
    This is the reading version of the Table class, which is used when retrieving a table.

    Args:
        name (str | None): Unique name of the table
        created_time (int | None): Time the table was created.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        name: str | None = None,
        database: str | None = None,
        created_time: int | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(name, database)
        self.created_time = created_time
        self._cognite_client = cast("CogniteClient", cognite_client)

        self._db_name: str | None = None

    def as_write(self) -> "RawTableWrite":
        """Returns this Table as a TableWrite"""
        if self.name is None or self.database is None:
            raise ValueError("name and database are required to create a Table")
        return RawTableWrite(name=self.name, database=self.database)


class RawTableWrite(RawTableCore):
    """A NoSQL database table to store customer data
    This is the writing version of the Table class, which is used when creating a table.

    Args:
        name (str): Unique name of the table
    """

    def __init__(
        self,
        name: str,
        database: str,
    ) -> None:
        super().__init__(name, database)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "RawTableWrite":
        return cls(resource["name"], resource["database"])

    def as_write(self) -> "RawTableWrite":
        """Returns this TableWrite instance."""
        return self


class RawTableWriteList(CogniteResourceList[RawTableWrite]):
    _RESOURCE = RawTableWrite

    def as_ids(self) -> list[RawTableID]:
        """Returns this TableWriteList as a list of TableIDs"""
        return [table.as_id() for table in self.data]


class RawTableList(
    WriteableCogniteResourceList[RawTableWrite, RawTable],
):
    _RESOURCE = RawTable

    def as_write(self) -> RawTableWriteList:
        """Returns this TableList as a TableWriteList"""
        return RawTableWriteList([table.as_write() for table in self.data])

    def as_ids(self) -> list[RawTableID]:
        """Returns this TableList as a list of TableIDs"""
        return [table.as_id() for table in self.data]



import re


def to_camel(string: str) -> str:
    """Convert snake_case_name to camelCaseName.

    Args:
        string: The string to convert.
    Returns:
        camelCase of the input string.

    Examples:
        >>> to_camel("a_b")
        'aB'
        >>> to_camel("ScenarioInstance_priceForecast")
        'scenarioInstancePriceForecast'
    """
    if "_" in string:
        # Could be a combination of snake and pascal/camel case
        parts = string.split("_")
        pascal_splits = [to_pascal(subpart) for part in parts for subpart in part.split("-") if subpart]
    elif "-" in string:
        # Could be a combination of kebab and pascal/camel case
        parts = string.split("-")
        pascal_splits = [to_pascal(subpart) for part in parts for subpart in part.split("_") if subpart]
    else:
        # Assume is pascal/camel case
        # Ensure pascal
        string = string[0].upper() + string[1:]
        pascal_splits = [string]
    string_split = []
    for part in pascal_splits:
        string_split.extend(re.findall(r"[A-Z][a-z0-9]*", part))
    if not string_split:
        string_split = [string]
    try:
        return string_split[0].casefold() + "".join(word.capitalize() for word in string_split[1:])
    except IndexError:
        return ""


def to_pascal(string: str) -> str:
    """Convert string to PascalCaseName.

    Args:
        string: The string to convert.
    Returns:
        PascalCase of the input string.

    Examples:
        >>> to_pascal("a_b")
        'AB'
        >>> to_pascal('camel_case')
        'CamelCase'
    """
    camel = to_camel(string)
    return f"{camel[0].upper()}{camel[1:]}" if camel else ""


def to_snake(string: str) -> str:
    """
    Convert input string to snake_case

    Args:
        string: The string to convert.
    Returns:
        snake_case of the input string.

    Examples:
        >>> to_snake("aB")
        'a_b'
        >>> to_snake('CamelCase')
        'camel_case'
        >>> to_snake('camelCamelCase')
        'camel_camel_case'
        >>> to_snake('Camel2Camel2Case')
        'camel_2_camel_2_case'
        >>> to_snake('getHTTPResponseCode')
        'get_http_response_code'
        >>> to_snake('get200HTTPResponseCode')
        'get_200_http_response_code'
        >>> to_snake('getHTTP200ResponseCode')
        'get_http_200_response_code'
        >>> to_snake('HTTPResponseCode')
        'http_response_code'
        >>> to_snake('ResponseHTTP')
        'response_http'
        >>> to_snake('ResponseHTTP2')
        'response_http_2'
        >>> to_snake('Fun?!awesome')
        'fun_awesome'
        >>> to_snake('Fun?!Awesome')
        'fun_awesome'
        >>> to_snake('10CoolDudes')
        '10_cool_dudes'
        >>> to_snake('20coolDudes')
        '20_cool_dudes'
    """
    pattern = re.compile(r"[A-Z]?[a-z]+|[A-Z]+(?=[A-Z][a-z]|\d|\W|$)|\d+")
    if "_" in string:
        words = [word for section in string.split("_") for word in pattern.findall(section)]
    else:
        words = pattern.findall(string)
    return "_".join(map(str.lower, words))



from pathlib import Path

from rdflib import URIRef

from cognite.neat.graph._shared import MIMETypes
from cognite.neat.graph.extractors._base import BaseExtractor


class RdfFileExtractor(BaseExtractor):
    """Extract data from RDF files into Neat.

    Args:
        filepath (Path): The path to the RDF file.
        mime_type (MIMETypes, optional): The MIME type of the RDF file. Defaults to "application/rdf+xml".
        base_uri (URIRef, optional): The base URI to use. Defaults to None.
    """

    def __init__(
        self,
        filepath: Path,
        mime_type: MIMETypes = "application/rdf+xml",
        base_uri: URIRef | None = None,
    ):
        self.filepath = filepath
        self.mime_type = mime_type
        self.base_uri = base_uri



from abc import ABC
from dataclasses import dataclass
from typing import Any, ClassVar

from cognite.client.data_classes import data_modeling as dm

from .base import NeatValidationError, ValidationWarning

__all__ = [
    "DMSSchemaError",
    "DMSSchemaWarning",
    "IncompleteSchemaError",
    "MissingSpaceError",
    "MissingContainerError",
    "MissingContainerPropertyError",
    "MissingViewError",
    "MissingParentViewError",
    "MissingSourceViewError",
    "MissingEdgeViewError",
    "DirectRelationMissingSourceWarning",
    "ViewModelVersionNotMatchingWarning",
    "ViewModelSpaceNotMatchingWarning",
    "ViewMapsToTooManyContainersWarning",
    "DuplicatedViewInDataModelError",
    "ContainerPropertyUsedMultipleTimesError",
    "EmptyContainerWarning",
    "UnsupportedConnectionWarning",
    "MultipleReferenceWarning",
    "HasDataFilterOnNoPropertiesViewWarning",
    "HasDataFilterAppliedToTooManyContainersWarning",
    "ReverseRelationMissingOtherSideWarning",
    "NodeTypeFilterOnParentViewWarning",
    "MissingViewInModelWarning",
    "ViewSizeWarning",
    "ChangingContainerError",
    "ChangingViewError",
]


@dataclass(frozen=True)
class DMSSchemaError(NeatValidationError, ABC): ...


@dataclass(frozen=True)
class DMSSchemaWarning(ValidationWarning, ABC): ...


@dataclass(frozen=True)
class ViewSizeWarning(DMSSchemaWarning):
    description = (
        "The number of properties in the {view} view is {count} which is more than "
        "the recommended limit of {limit} properties. This can lead to performance issues."
    )
    fix = "Reduce the size of the view"
    error_name: ClassVar[str] = "ViewSizeWarning"

    view_id: dm.ViewId
    limit: int
    count: int

    def message(self) -> str:
        return self.description.format(view=repr(self.view_id), count=self.count, limit=self.limit)

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["limit"] = self.limit
        output["count"] = self.count
        return output


@dataclass(frozen=True)
class IncompleteSchemaError(DMSSchemaError):
    description = "This error is raised when the schema is claimed to be complete but missing some components"
    fix = "Either provide the missing components or change the schema to partial"
    missing_component: dm.ContainerId | dm.ViewId

    def message(self) -> str:
        return (
            "The data model schema is set to be complete, however, "
            f"the referred component {self.missing_component} is not preset."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["missing_component"] = self.missing_component
        return output


@dataclass(frozen=True)
class MissingSpaceError(DMSSchemaError):
    description = "The spaced referred to by the Container/View/Node/Edge/DataModel does not exist"
    fix = "Create the space"
    space: str
    referred_by: dm.ContainerId | dm.ViewId | dm.NodeId | dm.EdgeId | dm.DataModelId

    def message(self) -> str:
        return f"The space {self.space} referred to by {self.referred_by} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["space"] = self.space
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class MissingContainerError(DMSSchemaError):
    description = "The container referred to by the View does not exist"
    fix = "Create the container"
    error_name: ClassVar[str] = "MissingContainer"
    container: dm.ContainerId
    referred_by: dm.ViewId | dm.ContainerId

    def message(self) -> str:
        return f"The container {self.container} referred to by {self.referred_by} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container"] = self.container
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class MissingContainerPropertyError(DMSSchemaError):
    description = "The property referred to by the View does not exist in the container"
    fix = "Create the property"
    error_name: ClassVar[str] = "MissingContainerProperty"
    container: dm.ContainerId
    property: str
    referred_by: dm.ViewId

    def message(self) -> str:
        return (
            f"The property {self.property} referred to by the container {self.container} "
            f"does not exist in {self.referred_by}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container"] = self.container
        output["property"] = self.property
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class MissingViewError(DMSSchemaError):
    description = "The view referred to by the View/DataModel does not exist"
    fix = "Create the view"
    error_name: ClassVar[str] = "MissingView"
    view: dm.ViewId
    referred_by: dm.DataModelId | dm.ViewId

    def message(self) -> str:
        return f"The view {self.view} referred to by {self.referred_by} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view"] = self.view
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class MissingParentViewError(MissingViewError):
    description = "The parent view referred to by the View does not exist"
    fix = "Create the parent view"
    error_name: ClassVar[str] = "MissingParentView"
    referred_by: dm.ViewId

    def message(self) -> str:
        return f"The parent view {self.view} referred to by {self.referred_by} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class MissingSourceViewError(MissingViewError):
    description = "The source view referred to by the View does not exist"
    fix = "Create the source view"
    error_name: ClassVar[str] = "MissingSourceView"
    property: str
    referred_by: dm.ViewId

    def message(self) -> str:
        return f"The source view {self.view} referred to by {self.referred_by}.{self.property} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["property"] = self.property
        return output


@dataclass(frozen=True)
class MissingEdgeViewError(MissingViewError):
    description = "The edge view referred to by the View does not exist"
    fix = "Create the edge view"
    error_name: ClassVar[str] = "MissingEdgeView"
    property: str
    referred_by: dm.ViewId

    def message(self) -> str:
        return f"The edge view {self.view} referred to by {self.referred_by}.{self.property} does not exist"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["property"] = self.property
        output["referred_by"] = self.referred_by
        return output


@dataclass(frozen=True)
class DuplicatedViewInDataModelError(DMSSchemaError):
    description = "The view is duplicated in the DataModel"
    fix = "Remove the duplicated view"
    error_name: ClassVar[str] = "DuplicatedViewInDataModel"
    referred_by: dm.DataModelId
    view: dm.ViewId

    def message(self) -> str:
        return f"The view {self.view} is duplicated in the DataModel {self.referred_by}"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["referred_by"] = self.referred_by
        output["view"] = self.view
        return output


@dataclass(frozen=True)
class DirectRelationMissingSourceWarning(DMSSchemaWarning):
    description = "The source view referred to by the DirectRelation does not exist"
    fix = "Create the source view"
    error_name: ClassVar[str] = "DirectRelationMissingSource"
    view_id: dm.ViewId
    property: str

    def message(self) -> str:
        return f"The source view referred to by '{self.view_id.external_id}.{self.property}' does not exist."

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id
        output["property"] = self.property
        return output


@dataclass(frozen=True)
class ViewModelVersionNotMatchingWarning(DMSSchemaWarning):
    description = "The view model version does not match the data model version"
    fix = "Update the view model version to match the data model version"
    error_name: ClassVar[str] = "ViewModelVersionNotMatching"
    view_ids: list[dm.ViewId]
    data_model_version: str

    def message(self) -> str:
        return (
            f"The version in the views {self.view_ids} does not match the version in the data model "
            f"{self.data_model_version}. This is not recommended as it easily leads to confusion and errors. "
            f"Views are very cheap and we recommend you update the version of the views to match the data"
            f" model version, irrespective of whether the views have changed or not."
            " The approach of having same version of model components as the model itself "
            " is a globally recognized data modeling practice."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = [view_id.dump() for view_id in self.view_ids]
        output["data_model_version"] = self.data_model_version
        return output


@dataclass(frozen=True)
class ViewModelSpaceNotMatchingWarning(DMSSchemaWarning):
    description = "The view model space does not match the data model space"
    fix = "Update the view model space to match the data model space"
    error_name: ClassVar[str] = "ViewModelSpaceNotMatching"
    view_ids: list[dm.ViewId]
    data_model_space: str

    def message(self) -> str:
        return (
            f"The space in the views {self.view_ids} does not match the space in the data model "
            f"{self.data_model_space}. This is not recommended as it easily leads to confusion and errors. "
            f"Views are very cheap and we recommend you always have views in the same space as the data model."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = [view_id.dump() for view_id in self.view_ids]
        output["data_model_space"] = self.data_model_space
        return output


@dataclass(frozen=True)
class ViewMapsToTooManyContainersWarning(DMSSchemaWarning):
    description = "The view maps to more than 10 containers which impacts read/write performance of data model"
    fix = "Try to have as few containers as possible to which the view maps to"
    error_name: ClassVar[str] = "ViewMapsToTooManyContainers"
    view_id: dm.ViewId
    container_ids: set[dm.ContainerId]

    def message(self) -> str:
        return (
            f"The view {self.view_id} maps to total of {len(self.container_ids)},."
            "Mapping to more than 10 containers is not recommended and can lead to poor performances."
            "Re-iterate the data model design to reduce the number of containers to which the view maps to."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["container_ids"] = [container_id.dump() for container_id in self.container_ids]
        return output


@dataclass(frozen=True)
class MissingViewInModelWarning(DMSSchemaWarning):
    description = "The data model contains view pointing to views not present in the data model"
    fix = "Add the view(s) to the data model"
    error_name: ClassVar[str] = "MissingViewInModel"
    data_model_id: dm.DataModelId
    view_ids: set[dm.ViewId]

    def message(self) -> str:
        return f"The view(s) {self.view_ids} are missing in the data model {self.data_model_id}"

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["data_model_id"] = self.data_model_id.dump()
        output["view_id"] = [view_id.dump() for view_id in self.view_ids]
        return output


@dataclass(frozen=True)
class ContainerPropertyUsedMultipleTimesError(DMSSchemaError):
    description = "The container property is used multiple times by the same view property"
    fix = "Use unique container properties for when mapping to the same container"
    error_name: ClassVar[str] = "ContainerPropertyUsedMultipleTimes"
    container: dm.ContainerId
    property: str
    referred_by: frozenset[tuple[dm.ViewId, str]]

    def message(self) -> str:
        return (
            f"The container property {self.property} of {self.container} is used multiple times "
            f"by the same view {self.referred_by}"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container"] = self.container
        output["property"] = self.property
        output["referred_by"] = sorted(self.referred_by)
        return output


@dataclass(frozen=True)
class ChangingContainerError(DMSSchemaError):
    description = "You are adding to an existing model. "
    fix = "Keep the container the same"
    error_name: ClassVar[str] = "ChangingContainerError"
    container_id: dm.ContainerId
    changed_properties: list[str] | None = None
    changed_attributes: list[str] | None = None

    def __post_init__(self):
        # Sorting for deterministic output
        if self.changed_properties:
            self.changed_properties.sort()
        if self.changed_attributes:
            self.changed_attributes.sort()

    def message(self) -> str:
        if self.changed_properties:
            changed = f" properties {self.changed_properties}."
        elif self.changed_attributes:
            changed = f" attributes {self.changed_attributes}."
        else:
            changed = "."
        return (
            f"The container {self.container_id} has changed{changed}"
            "When extending model with extension set to addition or reshape, the container "
            "properties must remain the same"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container_id"] = self.container_id.dump()
        output["changed_properties"] = self.changed_properties
        return output


@dataclass(frozen=True)
class ChangingViewError(DMSSchemaError):
    description = "You are adding to an existing model. "
    fix = "Keep the view the same"
    error_name: ClassVar[str] = "ChangingViewError"
    view_id: dm.ViewId
    changed_properties: list[str] | None = None
    changed_attributes: list[str] | None = None

    def __post_init__(self):
        # Sorting for deterministic output
        if self.changed_properties:
            self.changed_properties.sort()
        if self.changed_attributes:
            self.changed_attributes.sort()

    def message(self) -> str:
        if self.changed_properties:
            changed = f" properties {self.changed_properties}."
        elif self.changed_attributes:
            changed = f" attributes {self.changed_attributes}."
        else:
            changed = "."

        return (
            f"The view {self.view_id} has changed{changed}"
            "When extending model with extension set to addition, the view properties must remain the same"
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["difference"] = self.changed_properties
        return output


@dataclass(frozen=True)
class EmptyContainerWarning(DMSSchemaWarning):
    description = "The container is empty"
    fix = "Add data to the container"
    error_name: ClassVar[str] = "EmptyContainerWarning"
    container_id: dm.ContainerId

    def message(self) -> str:
        return (
            f"The container {self.container_id} is empty. Is this intended? Skipping this container "
            "in the data model."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["container_id"] = self.container_id.dump()
        return output


@dataclass(frozen=True)
class UnsupportedConnectionWarning(DMSSchemaWarning):
    description = "The connection type is not supported by neat"
    fix = "Change the connection to a supported type"
    error_name: ClassVar[str] = "UnsupportedConnectionWarning"
    view_id: dm.ViewId
    property: str
    connection: str

    def message(self) -> str:
        return (
            f"The connection {self.connection} in {self.view_id}.{self.property} is not supported."
            "This property will be ignored."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["property"] = self.property
        output["connection"] = self.connection
        return output


@dataclass(frozen=True)
class ReverseRelationMissingOtherSideWarning(DMSSchemaWarning):
    description = "The relation is missing the other side"
    fix = "Add the other side of the relation"
    error_name: ClassVar[str] = "ReverseRelationMissingOtherSideWarning"
    view_id: dm.ViewId
    property: str

    def message(self) -> str:
        return f"The reverse relation specified in {self.view_id}.{self.property} is missing the other side."

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["property"] = self.property
        return output


@dataclass(frozen=True)
class MultipleReferenceWarning(DMSSchemaWarning):
    description = "The view is implements multiple views from other spaces"
    fix = "Neat expects maximum one implementation of a view from another space"
    error_name: ClassVar[str] = "MultipleReferenceWarning"
    view_id: dm.ViewId
    implements: list[dm.ViewId]

    def message(self) -> str:
        return f"The view {self.view_id} implements multiple views from other spaces: {self.implements}. " + self.fix

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["implements"] = [view.dump() for view in self.implements]
        return output


@dataclass(frozen=True)
class HasDataFilterOnNoPropertiesViewWarning(DMSSchemaWarning):
    description = "Attempting to set a HasData filter on a view without properties."
    fix = "Add properties to the view or use a node type filter"
    error_name: ClassVar[str] = "HasDataFilterOnNoPropertiesViewWarning"
    view_id: dm.ViewId

    def message(self) -> str:
        return (
            f"Cannot set hasData filter on view {self.view_id} as it does not have properties in any containers. "
            "Using a node type filter instead."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        return output


@dataclass(frozen=True)
class HasDataFilterAppliedToTooManyContainersWarning(DMSSchemaWarning):
    description = "The view filter hasData applied to more than 10 containers this will cause DMS API Error"
    fix = "Do not map to more than 10 containers, alternatively override the filter by using rawFilter"
    error_name: ClassVar[str] = "HasDataFilterAppliedToTooManyContainers"
    view_id: dm.ViewId
    container_ids: set[dm.ContainerId]

    def message(self) -> str:
        return (
            f"The view {self.view_id} HasData filter applied to total of {len(self.container_ids)},."
            "Applying HasData filter to more than 10 containers is not recommended and can lead to DMS API error."
            "Re-iterate the data model design to reduce the number of containers to which the view maps to."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["container_ids"] = [container_id.dump() for container_id in self.container_ids]
        return output


@dataclass(frozen=True)
class RawFilterAppliedToViewWarning(DMSSchemaWarning):
    description = "Raw filter is applied to a view, which is against the neat data modeling lifecycle."
    fix = "Do not use raw filter, instead use HasData filter or nodeType filter or change the data model design."
    error_name: ClassVar[str] = "RawFilterAppliedToView"
    view_id: dm.ViewId

    def message(self) -> str:
        return (
            f"RawFilter applied to the view {self.view_id}."
            " The usage of RawFilter is against the neat team recommendations and the neat data modeling lifecycle."
            " When opting for raw filter, the user is responsible for any errors that arise in neat."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        return output


@dataclass(frozen=True)
class NodeTypeFilterOnParentViewWarning(DMSSchemaWarning):
    description = (
        "Setting a node type filter on a parent view. This is not "
        "recommended as parent views are typically used for multiple type of nodes."
    )
    fix = "Use a HasData filter instead"
    error_name: ClassVar[str] = "NodeTypeFilterOnParentViewWarning"
    view_id: dm.ViewId

    def message(self) -> str:
        return (
            f"Setting a node type filter on parent view {self.view_id}. This is not recommended as "
            "parent views are typically used for multiple types of nodes."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        return output


@dataclass(frozen=True)
class HasDataFilterOnViewWithReferencesWarning(DMSSchemaWarning):
    description = (
        "Setting a hasData filter on a solution view which reference other containers is not recommended."
        "This will lead to no nodes being returned when querying the solution view."
    )
    fix = "Use a node type filter instead"
    error_name: ClassVar[str] = "HasDataFilterOnReferencedViewWarning"

    view_id: dm.ViewId
    references: list[dm.ViewId]

    def message(self) -> str:
        return (
            f"Setting a hasData filter on view {self.view_id} which references other views {self.references}. "
            "This is not recommended as it will lead to no nodes being returned when querying the solution view."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["view_id"] = self.view_id.dump()
        output["references"] = [view.dump() for view in sorted(self.references, key=lambda x: x.as_tuple())]
        return output


@dataclass(frozen=True)
class OtherDataModelsInSpaceWarning(DMSSchemaWarning):
    description = "The space contains other data models"
    fix = "Move the data models to their respective spaces."
    error_name: ClassVar[str] = "OtherDataModelsInSpaceWarning"
    space: str
    data_models: list[dm.DataModelId]

    def message(self) -> str:
        return (
            f"The space {self.space} contains data models from other spaces: {self.data_models}. {self.fix}"
            " It is recommended to only have one data model per space. This avoid potential conflicts and "
            "makes it easier to manage the data models."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["space"] = self.space
        output["data_models"] = [data_model.dump() for data_model in self.data_models]
        return output


@dataclass(frozen=True)
class SolutionOnTopOfSolutionModelWarning(DMSSchemaWarning):
    description = "The data model is a solution on top of another solution"
    fix = "Use the base solution as the data model"
    error_name: ClassVar[str] = "SolutionOnTopOfSolutionModelWarning"
    data_model: dm.DataModelId
    base_data_model: dm.DataModelId

    def message(self) -> str:
        return (
            f"The data model {self.data_model} is a solution model on top of another solution {self.base_data_model} "
            "model. This is not recommended as it can lead to confusion and errors. It is very hard to "
            "maintain a nested structure of solution models. Instead, only build solution models on "
            "top of enterprise models."
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["data_model"] = self.data_model.dump()
        output["base_data_model"] = self.base_data_model.dump()
        return output



"""This module performs importing of graph to TransformationRules pydantic class.
In more details, it traverses the graph and abstracts class and properties, basically
generating a list of rules based on which nodes that form the graph are made.
"""

from collections import UserDict, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, cast, overload

import pandas as pd
from pandas import ExcelFile

from cognite.neat.rules import issues
from cognite.neat.rules.issues import IssueList
from cognite.neat.rules.models import (
    RULES_PER_ROLE,
    AssetRules,
    DMSRules,
    DomainRules,
    InformationRules,
    RoleTypes,
    SchemaCompleteness,
)
from cognite.neat.rules.models.asset import AssetRulesInput
from cognite.neat.rules.models.dms import DMSRulesInput
from cognite.neat.rules.models.information import InformationRulesInput
from cognite.neat.utils.auxiliary import local_import
from cognite.neat.utils.spreadsheet import SpreadsheetRead, read_individual_sheet

from ._base import BaseImporter, Rules, _handle_issues

SOURCE_SHEET__TARGET_FIELD__HEADERS = [
    (
        "Properties",
        "Properties",
        {
            RoleTypes.domain_expert: "Property",
            RoleTypes.information_architect: "Property",
            RoleTypes.asset_architect: "Property",
            RoleTypes.dms_architect: "View Property",
        },
    ),
    ("Classes", "Classes", "Class"),
    ("Containers", "Containers", "Container"),
    ("Views", "Views", "View"),
]

MANDATORY_SHEETS_BY_ROLE: dict[RoleTypes, set[str]] = {
    role_type: {str(sheet_name) for sheet_name in RULES_PER_ROLE[role_type].mandatory_fields(use_alias=True)}
    for role_type in RoleTypes.__members__.values()
}


class MetadataRaw(UserDict):
    @classmethod
    def from_excel(cls, excel_file: ExcelFile, metadata_sheet_name: str) -> "MetadataRaw":
        return cls(pd.read_excel(excel_file, metadata_sheet_name, header=None).replace(float("nan"), None).values)

    @property
    def has_role_field(self) -> bool:
        return self.get("role") in [role.value for role in RoleTypes.__members__.values()]

    @property
    def role(self) -> RoleTypes:
        return RoleTypes(self["role"])

    @property
    def has_schema_field(self) -> bool:
        return self.get("schema") in [schema.value for schema in SchemaCompleteness.__members__.values()]

    @property
    def schema(self) -> SchemaCompleteness | None:
        if not self.has_schema_field:
            return None
        return SchemaCompleteness(self["schema"])

    def is_valid(self, issue_list: IssueList, filepath: Path) -> bool:
        if not self.has_role_field:
            issue_list.append(issues.spreadsheet_file.RoleMissingOrUnsupportedError(filepath))
            return False

        # check if there is a schema field if role is not domain expert
        if self.role != RoleTypes.domain_expert and not self.has_schema_field:
            issue_list.append(issues.spreadsheet_file.SchemaMissingOrUnsupportedError(filepath))
            return False
        return True


@dataclass
class ReadResult:
    sheets: dict[str, dict | list]
    read_info_by_sheet: dict[str, SpreadsheetRead]
    metadata: MetadataRaw

    @property
    def role(self) -> RoleTypes:
        return self.metadata.role

    @property
    def schema(self) -> SchemaCompleteness | None:
        return self.metadata.schema


class SpreadsheetReader:
    def __init__(
        self,
        issue_list: IssueList,
        required: bool = True,
        metadata: MetadataRaw | None = None,
        sheet_prefix: Literal["", "Last", "Ref"] = "",
    ):
        self.issue_list = issue_list
        self.required = required
        self.metadata = metadata
        self._sheet_prefix = sheet_prefix
        self._seen_files: set[Path] = set()
        self._seen_sheets: set[str] = set()

    @property
    def metadata_sheet_name(self) -> str:
        return f"{self._sheet_prefix}Metadata"

    @property
    def seen_sheets(self) -> set[str]:
        if not self._seen_files:
            raise ValueError("No files have been read yet.")
        return self._seen_sheets

    def sheet_names(self, role: RoleTypes) -> set[str]:
        names = MANDATORY_SHEETS_BY_ROLE[role]
        return {f"{self._sheet_prefix}{sheet_name}" for sheet_name in names if sheet_name != "Metadata"}

    def read(self, filepath: Path) -> None | ReadResult:
        with pd.ExcelFile(filepath) as excel_file:
            self._seen_files.add(filepath)
            self._seen_sheets.update(map(str, excel_file.sheet_names))
            metadata: MetadataRaw | None
            if self.metadata is not None:
                metadata = self.metadata
            else:
                metadata = self._read_metadata(excel_file, filepath)
                if metadata is None:
                    # The reading of metadata failed, so we can't continue
                    return None

            sheets, read_info_by_sheet = self._read_sheets(excel_file, metadata.role)
            if sheets is None or self.issue_list.has_errors:
                return None
            sheets["Metadata"] = dict(metadata)

            return ReadResult(sheets, read_info_by_sheet, metadata)

    def _read_metadata(self, excel_file: ExcelFile, filepath: Path) -> MetadataRaw | None:
        if self.metadata_sheet_name not in excel_file.sheet_names:
            if self.required:
                self.issue_list.append(
                    issues.spreadsheet_file.MetadataSheetMissingOrFailedError(
                        filepath, sheet_name=self.metadata_sheet_name
                    )
                )
            return None

        metadata = MetadataRaw.from_excel(excel_file, self.metadata_sheet_name)

        if not metadata.is_valid(self.issue_list, filepath):
            return None
        return metadata

    def _read_sheets(
        self, excel_file: ExcelFile, read_role: RoleTypes
    ) -> tuple[dict[str, dict | list] | None, dict[str, SpreadsheetRead]]:
        read_info_by_sheet: dict[str, SpreadsheetRead] = defaultdict(SpreadsheetRead)

        sheets: dict[str, dict | list] = {}

        expected_sheet_names = self.sheet_names(read_role)

        if missing_sheets := expected_sheet_names.difference(set(excel_file.sheet_names)):
            if self.required:
                self.issue_list.append(
                    issues.spreadsheet_file.SheetMissingError(cast(Path, excel_file.io), list(missing_sheets))
                )
            return None, read_info_by_sheet

        for source_sheet_name, target_sheet_name, headers_input in SOURCE_SHEET__TARGET_FIELD__HEADERS:
            source_sheet_name = f"{self._sheet_prefix}{source_sheet_name}"

            if source_sheet_name not in excel_file.sheet_names:
                continue
            if isinstance(headers_input, dict):
                headers = headers_input[read_role]
            else:
                headers = headers_input

            try:
                sheets[target_sheet_name], read_info_by_sheet[source_sheet_name] = read_individual_sheet(
                    excel_file, source_sheet_name, return_read_info=True, expected_headers=[headers]
                )
            except Exception as e:
                self.issue_list.append(issues.spreadsheet_file.ReadSpreadsheetsError(cast(Path, excel_file.io), str(e)))
                continue

        return sheets, read_info_by_sheet


class ExcelImporter(BaseImporter):
    """Import rules from an Excel file.

    Args:
        filepath (Path): The path to the Excel file.
    """

    def __init__(self, filepath: Path):
        self.filepath = filepath

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules: ...

    @overload
    def to_rules(
        self,
        errors: Literal["continue"] = "continue",
        role: RoleTypes | None = None,
    ) -> tuple[Rules | None, IssueList]: ...

    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        issue_list = IssueList(title=f"'{self.filepath.name}'")
        if not self.filepath.exists():
            issue_list.append(issues.spreadsheet_file.SpreadsheetNotFoundError(self.filepath))
            return self._return_or_raise(issue_list, errors)

        user_reader = SpreadsheetReader(issue_list)
        user_read = user_reader.read(self.filepath)
        if user_read is None or issue_list.has_errors:
            return self._return_or_raise(issue_list, errors)

        last_read: ReadResult | None = None
        if any(sheet_name.startswith("Last") for sheet_name in user_reader.seen_sheets):
            last_read = SpreadsheetReader(issue_list, required=False, sheet_prefix="Last").read(self.filepath)
        reference_read: ReadResult | None = None
        if any(sheet_name.startswith("Ref") for sheet_name in user_reader.seen_sheets):
            reference_read = SpreadsheetReader(issue_list, sheet_prefix="Ref").read(self.filepath)

        if issue_list.has_errors:
            return self._return_or_raise(issue_list, errors)

        if reference_read and user_read.role != reference_read.role:
            issue_list.append(issues.spreadsheet_file.RoleMismatchError(self.filepath))
            return self._return_or_raise(issue_list, errors)

        sheets = user_read.sheets
        original_role = user_read.role
        read_info_by_sheet = user_read.read_info_by_sheet
        if last_read:
            sheets["last"] = last_read.sheets
            read_info_by_sheet.update(last_read.read_info_by_sheet)
            if reference_read:
                # The last rules will also be validated against the reference rules
                sheets["last"]["reference"] = reference_read.sheets  # type: ignore[call-overload]
        if reference_read:
            sheets["reference"] = reference_read.sheets
            read_info_by_sheet.update(reference_read.read_info_by_sheet)

        rules_cls = RULES_PER_ROLE[original_role]
        with _handle_issues(
            issue_list,
            error_cls=issues.spreadsheet.InvalidSheetError,
            error_args={"read_info_by_sheet": read_info_by_sheet},
        ) as future:
            rules: Rules
            if rules_cls is DMSRules:
                rules = DMSRulesInput.load(sheets).as_rules()
            elif rules_cls is InformationRules:
                rules = InformationRulesInput.load(sheets).as_rules()
            elif rules_cls is AssetRules:
                rules = AssetRulesInput.load(sheets).as_rules()
            else:
                rules = rules_cls.model_validate(sheets)  # type: ignore[attr-defined]

        if future.result == "failure" or issue_list.has_errors:
            return self._return_or_raise(issue_list, errors)

        return self._to_output(
            rules,
            issue_list,
            errors=errors,
            role=role,
        )


class GoogleSheetImporter(BaseImporter):
    """Import rules from a Google Sheet.

    .. warning::

        This importer is experimental and may not work as expected.

    Args:
        sheet_id (str): The Google Sheet ID.
        skiprows (int): The number of rows to skip when reading the Google Sheet.
    """

    def __init__(self, sheet_id: str, skiprows: int = 1):
        self.sheet_id = sheet_id
        self.skiprows = skiprows

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None) -> Rules: ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList]: ...

    def to_rules(
        self, errors: Literal["raise", "continue"] = "continue", role: RoleTypes | None = None
    ) -> tuple[Rules | None, IssueList] | Rules:
        local_import("gspread", "google")
        import gspread  # type: ignore[import]

        role = role or RoleTypes.domain_expert
        rules_model = cast(DomainRules | InformationRules | AssetRules | DMSRules, RULES_PER_ROLE[role])

        client_google = gspread.service_account()
        google_sheet = client_google.open_by_key(self.sheet_id)
        sheets = {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in google_sheet.worksheets()}
        sheet_names = {str(name).lower() for name in sheets.keys()}

        if missing_sheets := rules_model.mandatory_fields().difference(sheet_names):
            raise ValueError(f"Missing mandatory sheets: {missing_sheets}")

        if role == RoleTypes.domain_expert:
            output = rules_model.model_validate(sheets)
        elif role == RoleTypes.information_architect:
            output = rules_model.model_validate(sheets)
        elif role == RoleTypes.dms_architect:
            output = rules_model.model_validate(sheets)
        else:
            raise ValueError(f"Role {role} is not valid.")

        return self._to_output(output, IssueList(), errors=errors, role=role)



import warnings
from typing import cast

from rdflib import RDF, Graph, URIRef
from rdflib.query import ResultRow

from cognite.neat.rules.models.entities import ClassEntity
from cognite.neat.rules.models.information import InformationRules
from cognite.neat.utils.utils import remove_namespace

from ._construct import build_construct_query


class Queries:
    """Helper class for storing standard queries for the graph store."""

    def __init__(self, graph: Graph, rules: InformationRules | None = None):
        self.graph = graph
        self.rules = rules

    def list_instances_ids_of_class(self, class_uri: URIRef, limit: int = -1) -> list[URIRef]:
        """Get instances ids for a given class

        Args:
            class_uri: Class for which instances are to be found
            limit: Max number of instances to return, by default -1 meaning all instances

        Returns:
            List of class instance URIs
        """
        query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .} LIMIT X".replace(
            "class", class_uri
        ).replace("LIMIT X", "" if limit == -1 else f"LIMIT {limit}")
        return [cast(tuple, res)[0] for res in list(self.graph.query(query_statement))]

    def list_instances_of_type(self, class_uri: URIRef) -> list[ResultRow]:
        """Get all triples for instances of a given class

        Args:
            class_uri: Class for which instances are to be found

        Returns:
            List of triples for instances of the given class
        """
        query = (
            f"SELECT ?instance ?prop ?value "
            f"WHERE {{ ?instance rdf:type <{class_uri}> . ?instance ?prop ?value . }} order by ?instance "
        )

        # Select queries gives an iterable of result rows
        return cast(list[ResultRow], list(self.graph.query(query)))

    def triples_of_type_instances(self, rdf_type: str) -> list[tuple[str, str, str]]:
        """Get all triples of a given type.

        This method assumes the graph has been transformed into the default namespace.
        """

        if self.rules:
            query = (
                f"SELECT ?instance ?prop ?value "
                f"WHERE {{ ?instance a <{self.rules.metadata.namespace[rdf_type]}> . ?instance ?prop ?value . }} "
                "order by ?instance"
            )

            result = self.graph.query(query)

            # We cannot include the RDF.type in case there is a neat:type property
            return [remove_namespace(*triple) for triple in result if triple[1] != RDF.type]  # type: ignore[misc, index]
        else:
            warnings.warn("No rules found for the graph store, returning empty list.", stacklevel=2)
            return []

    def construct_instances_of_class(self, class_: str, properties_optional: bool = True) -> list[tuple[str, str, str]]:
        """CONSTRUCT instances for a given class from the graph store

        Args:
            class_: Class entity for which we want to generate query
            properties_optional: Whether to make all properties optional, default True

        Returns:
            List of triples for instances of the given class
        """

        if self.rules and (
            query := build_construct_query(
                ClassEntity(prefix=self.rules.metadata.prefix, suffix=class_),
                self.graph,
                self.rules,
                properties_optional,
            )
        ):
            result = self.graph.query(query)

            # We cannot include the RDF.type in case there is a neat:type property
            return [remove_namespace(*triple) for triple in result if triple[1] != RDF.type]  # type: ignore[misc, index]
        else:
            warnings.warn("No rules found for the graph store, returning empty list.", stacklevel=2)
            return []

    def list_triples(self, limit: int = 25) -> list[ResultRow]:
        """List triples in the graph store

        Args:
            limit: Max number of triples to return, by default 25

        Returns:
            List of triples
        """
        query = f"SELECT ?subject ?predicate ?object WHERE {{ ?subject ?predicate ?object }} LIMIT {limit}"
        return cast(list[ResultRow], list(self.graph.query(query)))

    def list_types(self, limit: int = 25) -> list[ResultRow]:
        """List types in the graph store

        Args:
            limit: Max number of types to return, by default 25

        Returns:
            List of types
        """
        query = f"SELECT DISTINCT ?type WHERE {{ ?subject a ?type }} LIMIT {limit}"
        return cast(list[ResultRow], list(self.graph.query(query)))
